import pydantic
from typing import List, Dict, Iterable, Optional, Any, Union, Callable
import collections
from abc import ABC, abstractmethod


"""---------------------Generator and Iterator Handler-----------------"""


# noinspection PyCompatibility
class GeneratorStorage(pydantic.BaseModel):
    """general object dataclass for generators"""
    generator_arr: Optional[Any] = collections.deque([])
    iterable_arr: Optional[List] = []


GeneratorStorage = GeneratorStorage()


def format_address(document: Dict) -> Union[str, None]:
    """format address based on US standards"""
    if document['unit'] == '':
        primary_address_search = document['number'] + ' ' + document['street'] + ', ' + document['region'] + ', ' + \
                                 document['postcode']
    else:
        primary_address_search = document['number'] + ' ' + document['street'] + ', ' + document['region'] + ', ' + \
                                 document['postcode'] + ', ' + document['unit']

    return primary_address_search


def clean_document(document: Dict) -> dict:
    """clean the document from elasticsearch or pymongo"""
    try:
        return document['_source']
    except KeyError:
        return document

"""------------------------Generator strategy handlers-----------------------"""


class GeneratorInternalStrategy(ABC):
    """abstract base class for generator/iterator strategy pattern"""
    @abstractmethod
    def generator_runner(self, iterable: Iterable, parser: Callable) -> List:
        """run generator"""
        pass


class RegularGenerator(GeneratorInternalStrategy):
    """ordinary generator algorithm"""

    def generator_runner(self, iterable: Iterable, parser: Callable) -> List:
        """run generator function"""

        # clear the generator storage first
        GeneratorStorage.generator_arr.clear()

        for document in iterable:
            # clean document
            document = clean_document(document)

            # coordinate parsing
            longitude, latitude = parser(document)

            # avoid blank street names
            if document['properties']['street'] != '' and document['properties']['number'] != '':
                address = format_address(document['properties'])
                micro_packet = {"address": address, "latitude": latitude, "longitude": longitude}

                # add to generator iterable
                GeneratorStorage.generator_arr.append(micro_packet)

            else:
                continue

        return GeneratorStorage.generator_arr


class SpecializedGeneratorPN32(GeneratorInternalStrategy):
    """Specialized Generator for other output types"""

    def generator_runner(self, iterable: Iterable, parser: Callable) -> List:
        """run specialized generator with diff. formatting"""

        # clear the generator first
        GeneratorStorage.generator_arr.clear()

        for packet in iterable:
            # clean packet
            packet = clean_document(packet)

            # make sure there are no blank streets
            if packet['properties']['street'] != '' and packet['properties']['number'] != '':

                # get coordinates from encapsulated parser callable
                latitude, longitude = parser(packet)

                # create the data packet with the pydantic model base
                data_packet = {'property_details': {'number': packet['properties']['number'],
                                                    'street': packet['properties']['street'],
                                                    'district': packet['properties']['district'],
                                                    'state': packet['properties']['region'],
                                                    'postcode': packet['properties']['postcode']},
                               'location': dict(zip(['latitude', 'longitude'], [latitude, longitude]))}

                # add to generator iterable
                GeneratorStorage.generator_arr.append(data_packet)

            else:
                continue

        return GeneratorStorage.generator_arr


class SpecializedGeneratorDN32(GeneratorInternalStrategy):
    """generator strategy for DN32 setting (density query)"""

    def generator_runner(self, iterable: Iterable, parser: Callable) -> List:
        """DN32 generator function"""

        # clear first
        GeneratorStorage.generator_arr.clear()

        for packet in iterable:
            # clean document from BSON blob
            packet = clean_document(packet)

            # append only coordinates for density search
            GeneratorStorage.generator_arr.append(parser(packet))

        return GeneratorStorage.generator_arr


"""iterator strategies here: regular and PR32 method"""


class IteratorInternalStrategies(ABC):
    """abstract base class for iterator strategy pattern"""
    @abstractmethod
    def iterator_parse(self, handler_obj: Any, parser: Callable) -> List:
        """base function for iterator strategies"""
        pass


class RegularIterator(IteratorInternalStrategies):
    """inherited subclass for regular iterator"""
    def iterator_parse(self, handler_obj: Any, parser: Callable) -> List:
        """regular iterator function"""

        # clear storages
        GeneratorStorage.iterable_arr.clear()
        GeneratorStorage.generator_arr.clear()

        for document in handler_obj:

            # clean document
            document = clean_document(document)

            # make sure there are no empty street or numbers
            if document['properties']['street'] != '' and document['properties']['number'] != '':
                # format the address
                address = format_address(document['properties'])

                # extract coordinates
                latitude, longitude = parser(document)

                # create data packet for append purposes
                micro_packet = {"address": address, "latitude": latitude, "longitude": longitude}

                # dataclass storage iterable_arr append method
                GeneratorStorage.iterable_arr.append(micro_packet)
            else:
                continue

        return GeneratorStorage.iterable_arr


class IteratorStrategyPR32(IteratorInternalStrategies):
    """specialized iterator method"""

    def iterator_parse(self, handler_obj: Any, parser: Callable) -> List:
        """PR32 iterator function"""

        # clear storages
        GeneratorStorage.iterable_arr.clear()
        GeneratorStorage.generator_arr.clear()

        for packet in handler_obj:
            # clean packet
            packet = clean_document(packet)

            # create data packet
            data_packet = {'property_details': {'number': packet['properties']['number'],
                                                'street': packet['properties']['street'],
                                                'district': packet['properties']['district'],
                                                'state': packet['properties']['region'],
                                                'postcode': packet['properties']['postcode']},
                           'location': dict(zip(['latitude', 'longitude'], parser(packet)))}

            GeneratorStorage.iterable_arr.append(data_packet)

        return GeneratorStorage.iterable_arr


"""----------------------------manage the generator and iterator methods------------------------------"""


class RunStrategy:
    """generator director class"""

    def __init__(self, parser, *args):
        self.parser = parser
        self.generator_strategy = args[0]
        self.iterator_strategy = args[1]

    def run_generator_internal(self, BSON_obj: Iterable) -> List:
        """run the generator, integrated strategy"""
        generator_output = self.generator_strategy.generator_runner(BSON_obj, self.parser)
        return generator_output

    def run_iterator_internal(self, BSON_obj: Iterable) -> List:
        """run iterator integrated strategy"""
        iterable_output = self.iterator_strategy.iterator_parse(BSON_obj, self.parser)
        return iterable_output


def run_generator(BSON_obj: Iterable, strategy, parser: Callable) -> List:
    """generator runner"""
    RG = RunStrategy(parser, strategy, None)
    return RG.run_generator_internal(BSON_obj)


def run_iterator(BSON_obj: Iterable, strategy, parser: Callable) -> List:
    """iterator runner"""
    RG = RunStrategy(parser, None, strategy)
    return RG.run_iterator_internal(BSON_obj)



"""create elasticsearch coordinate set for search"""

cimport cython

ctypedef fused value_type:
    cython.int
    cython.double
    cython.str

# noinspection PyMissingOrEmptyDocstring
cpdef list coordinate_zipped(list latitudes, list longitudes, bint reverse_param):

    """Cython/C function for handling long list of zipped coordinates"""

    cdef list coordinates = []
    cdef int i = 0
    cdef str coordinate = ''

    if reverse_param:
        for i in range(len(latitudes)):
            coordinate = '{}, {}'.format(latitudes[i],longitudes[i])
            coordinates.append(coordinate)
    else:
        for i in range(len(latitudes)):
            coordinate = '{}, {}'.format(longitudes[i], latitudes[i])
            coordinates.append(coordinate)

    return coordinates





cimport cython

ctypedef fused my_fused_type:
    cython.list
    cython.str

# noinspection PyTypeChecker
cpdef list is_duplicates(list edges_arr):
    """
    :param edges_arr: 
    :return: bint duplicate + exact dup edges
    """

    cdef list duplicates = []
    cdef list i = []
    cdef list item = []
    cdef bint dup_counter = False

    cdef list refined_edges = [str(i) for i in edges_arr]

    for edge in refined_edges:
        if refined_edges.count(edge) > 1:
            dup_counter = True
            duplicates.append(edge)

    if dup_counter:
        return [True, duplicates]
    else:
        return [False, duplicates]



cpdef create_packet(dict address_components, str formatted_address, list coordinate, str certainty, float accuracy, bint address_val):
    """
    :param address_val:
    :param accuracy:
    :param certainty:
    :param address_components:
    :param formatted_address: 
    :param coordinate: 
    :return: 
    """

    cdef dict inner_data_packet = {}

    inner_data_packet = {'address_parts': address_components,
                         'certainty': certainty,
                         'coordinate': coordinate,
                         'formatted_address': formatted_address,
                         'accuracy': accuracy,
                         'address_validation': address_val}

    return inner_data_packet


import time
import sys
import pymongo
import ssl
import collections
from typing import Callable, Any
sys.path.append('/places_geocode')
sys.path.append('/places_geocode/app/models')
from places_geocode.app.models.generator_model import *


class HandleArray:
    """handle iterable types with iterable/generator decision tree"""

    def __init__(self, array):
        self.array = array
        self.generator_arr = collections.deque([])
        self.iterable_arr = []

    def __repr__(self):
        return str(self.iterable_arr)

    def __len__(self):
        return len(self.array)

    def __iter__(self):
        return iter(self.array)

    def __add__(self, arr, **kwargs):
        if len(kwargs) > 0:
            if kwargs['option'] == 'GEN':
                if isinstance(arr, list) or isinstance(arr, dict):
                    self.generator_arr.append(arr)
                else:
                    raise Exception('Error Occurred in Generator BSON conversion')
            else:
                if isinstance(arr, list) or isinstance(arr, dict):
                    self.iterable_arr.append(arr)
                else:
                    raise Exception('Error Occurred in Iterator BSON conversion')
        else:
            if isinstance(arr, list) or isinstance(arr, dict):
                self.iterable_arr.append(arr)
            else:
                raise Exception('Error Occurred in Iterator BSON conversion')

    def __get__(self):
        return self.iterable_arr

    def get_iter_array(self):
        """return the iterable array"""
        return self.iterable_arr

    @staticmethod
    def _generator(array):
        """generator core internals"""
        for doc in array:
            yield doc

    @staticmethod
    def format_address(document):
        """format the address returned from iterator or generator"""
        document = document['properties']
        if document['unit'] == '':
            primary_address_search = document['number'] + ' ' + document['street'] + ', ' + document['region'] + ', ' + \
                                     document['postcode']
        else:
            primary_address_search = document['number'] + ' ' + document['street'] + ', ' + document['region'] + ', ' + \
                                     document['postcode'] + ', ' + document['unit']

        return primary_address_search

    def run_generator(self, strategy: Callable, parser: Callable):
        """run generator function"""

        # create the BSON-generator object
        BSON_obj = self._generator(self.array)

        # run the generator model based on the strategies in generator_models.py
        return run_generator(BSON_obj, strategy, parser)

    def run_iterator(self, strategy: Callable, parser: Callable):
        """run iterator function, considering strategy chosen"""
        return run_iterator(self.array, strategy, parser)


def handler_func(array, strategy: Any, parser: Callable, **kwargs):
    """director function for the generator algorithms"""

    # start the timer for runtime analysis
    start = time.perf_counter()

    # init the handler
    handler = HandleArray(array)

    try:
        # generator
        if kwargs['option'] == 0:
            # run handler generator function with current strategy
            return handler.run_generator(strategy, parser)

        # iterator
        elif kwargs['option'] == 1:
            # run iterator
            return handler.run_iterator(strategy, parser)

    except AttributeError:
        raise Exception("Incompatible Strategy Chosen")


# testing function for the generator functionality
def test_function():
    client = pymongo.MongoClient(
        'mongodb+srv://martin_mashalov:Barca2004!@locationpoints.wi1dk.mongodb.net/test', maxPoolSize=125,
        ssl_cert_reqs=ssl.CERT_NONE)
    db = client['PeekDB']
    geoCoords = db['propLocations']

    query = geoCoords.find({})[:180]

    try:
        return handler_func(query, RegularIterator(), option=0)
    except AttributeError:
        raise Exception("Incompatible Strategy Chosen")



# noinspection PyMissingOrEmptyDocstring
cpdef list coordinate_zipped(list latitudes, list longitudes, bint reverse_param):

    """Cython/C function for handling long list of zipped coordinates"""

    cdef float lat = 0.0
    cdef float long = 0.0
    cdef list coordinates = []
    cdef int i = 0

    if reverse_param:
        for i in range(len(latitudes)):
            coordinates.append([longitudes[i], latitudes[i]])
    else:
        for i in range(len(latitudes)):
            coordinates.append([latitudes[i], longitudes[i]])

    return coordinates





import requests
import time
import pymongo
import ssl
import json

from sklearn.feature_extraction.text import TfidfVectorizer
import numpy as np
from scipy.spatial import ConvexHull
from difflib import SequenceMatcher
from numba import njit
from CythonOperations import zip_coordinate, TFIDF_max, null_parameters


def test_endpoints():
    token = 'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiJtYXJ0aW4ubWFzaGFsb3YiLCJleHAiOjE2MjU5ODI1NjV9.OPugeH0picyrl9UrvWudv15S3tUBezZJQqOtdhSKuTY'
    my_headers = {'Authorization': token,
                  'runtime': 'X-Process-Time',
                  'accept': 'application/json'}
    url = 'http://127.0.0.1:8000/load_properties_api/parcel?latitude=43.0961466&longitude=-77.6337776&radius=100&username=martin.mashalov&reverse_param=false'
    start = time.perf_counter()
    data = requests.get(url,
                        headers=my_headers)

    print(data.json())
    print(time.perf_counter() - start)


def create_test_file_coordinates():
    client = pymongo.MongoClient(
        'mongodb+srv://martin_mashalov:Barca2004!@locationpoints.wi1dk.mongodb.net/test',
        ssl_cert_reqs=ssl.CERT_NONE)
    db = client['PeekDB']

    geoCoords = db['propLocations']
    json_structure = {'coordinates': []}

    with open('../analytics/sample_data/test_file_coordinates.json', 'a+') as file:
        for doc in geoCoords.find({})[0:20]:
            coordinates = doc['location']['coordinates']
            json_structure['coordinates'].append(coordinates)
        print(json_structure)
        json.dump(json_structure, file)


# create_test_file_coordinates()

def create_test_file_addresses():
    client = pymongo.MongoClient(
        'mongodb+srv://martin_mashalov:Barca2004!@locationpoints.wi1dk.mongodb.net/test',
        ssl_cert_reqs=ssl.CERT_NONE)
    db = client['PeekDB']

    geoCoords = db['propLocations']
    json_data = {'addresses': []}

    with open('test_file_addresses.json', 'a+') as file:
        for doc in geoCoords.find({})[0:20]:
            number = doc['properties']['number']
            street = doc['properties']['street']
            city = doc['properties']['city']
            postcode = doc['properties']['postcode']
            region = doc['properties']['region']
            unit = doc['properties']['unit']

            if unit != '':
                address = f'{number} {street}, {city}, {region}, {postcode}, {unit}'
            else:
                address = f'{number} {street}, {city}, {region}, {postcode}'

            json_data['addresses'].append(address)
        print(json_data)
        json.dump(json_data, file)


# create_test_file_addresses()

def create_test_spreadsheet():
    client = pymongo.MongoClient(
        'mongodb+srv://martin_mashalov:Barca2004!@locationpoints.wi1dk.mongodb.net/test',
        ssl_cert_reqs=ssl.CERT_NONE)
    db = client['PeekDB']

    geoCoords = db['propLocations']
    json_data = {'addresses': []}

    with open('../analytics/sample_data/test_spreadsheet_address.json', 'a+') as file:
        for doc in geoCoords.find({})[:100]:
            number = doc['properties']['number']
            street = doc['properties']['street']
            city = doc['properties']['city']
            postcode = doc['properties']['postcode']
            region = doc['properties']['region']
            unit = doc['properties']['unit']

            if unit != '':
                address = f'{number} {street}, {city}, {region}, {postcode}, {unit}'
            else:
                address = f'{number} {street}, {city}, {region}, {postcode}'

            json_data['addresses'].append(address)
        json.dump(json_data, file)


# create_test_spreadsheet()
def create_coordinates_spreadsheet():
    client = pymongo.MongoClient(
        'mongodb+srv://martin_mashalov:Barca2004!@locationpoints.wi1dk.mongodb.net/test',
        ssl_cert_reqs=ssl.CERT_NONE)
    db = client['PeekDB']

    geoCoords = db['propLocations']
    json_data = {'coordinates': []}

    with open('../analytics/sample_data/test_spreadsheet_coordinates.json', 'a+') as file:
        for doc in geoCoords.find({})[:14000]:
            coordinate = doc['location']['coordinates']

            json_data['coordinates'].append(coordinate)
        json.dump(json_data, file)


# create_coordinates_spreadsheet()

client = pymongo.MongoClient(
    'mongodb+srv://martin_mashalov:Barca2004!@locationpoints.wi1dk.mongodb.net/test',
    ssl_cert_reqs=ssl.CERT_NONE)
db = client['PeekDB']
geoCoords = db['propLocations']


def find_similarity(street, street_comparison) -> float:
    return SequenceMatcher(None, street.lower(), street_comparison['properties']['street'].lower()).ratio()


def polygon_search():
    # edges = [[[73.705034, 41.504346], [73.820142, 41.509602], [73.724686, 41.460176], [73.705034, 41.504346]]]
    edges = [[[41.010639, 73.860892], [41.010777, 73.862245], [41.011537, 73.862129], [41.010639, 73.860892]]]
    # new_edges = [[[41.010639, 73.860892], [41.010777, 73.862245], [41.011537, 73.862129]]]

    start = time.perf_counter()
    query = geoCoords.find({"location": {"$geoWithin": {"$geometry": {"type": "Polygon", "coordinates": edges}}}})
    print(query.count())
    print(time.perf_counter() - start)


# polygon_search()


def query():
    """test query function"""
    result = geoCoords.find(
        {'$and': [{'properties.number': str(20)}, {'properties.postcode': str(10522)}]})
    return result


def test_similarity():
    """test similarity func with TFIDF algorithm"""

    street_query = "Manor House"

    result = query()

    current_max = 0
    current_max_result = None

    for street_comparison in result:
        similarity = SequenceMatcher(None, street_query.lower(),
                                     street_comparison['properties']['street'].lower()).ratio()

        if similarity > current_max:
            current_max = similarity
            current_max_result = street_comparison
        else:
            continue

    return [current_max, current_max_result, current_max_result['properties']['street']]


"""start1 = time.perf_counter()
print(TFIDF_max.find_max_similarity("Manor House", query()))
print(time.perf_counter() - start1)

start2 = time.perf_counter()
print(test_similarity())
print(time.perf_counter() - start2)"""


def create_coordinate_test(lats: list, longs: list, reverse_param: bool):
    """
    :param lats:
    :param longs:
    :param reverse_param:
    """

    if reverse_param:
        return [[long, lat] for lat, long in zip(lats, longs)]
    else:
        return [[lat, long] for lat, long in zip(lats, longs)]


"""start = time.perf_counter()
create_coordinate_test([43.0900963, 43.0913689, 43.0920867], [-77.6463877, -77.6467333, -77.6439961], False)
first_time = time.perf_counter() - start

start2 = time.perf_counter()
zip_coordinate.coordinate_zipped([43.0900963, 43.0913689, 43.0920867, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3], [-77.6463877, -77.6467333, -77.6439961, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3], False)
print(first_time/(time.perf_counter() - start2))"""


def boolean_test1():
    """
    return: boolean test
    """
    x = 5
    print(not not x == 5)


def boolean_test2():
    """
    return: boolean test
    """
    x = 5
    print(bool(x == 5))


"""not_array = []
bool_array = []

for _ in range(100000):
    start1 = time.perf_counter()
    boolean_test1()
    not_time = time.perf_counter() - start1
    not_array.append(not_time)

    start2 = time.perf_counter()
    boolean_test2()
    bool_time = time.perf_counter() - start2
    bool_array.append(bool_time)

print(sum(not_array), "----", sum(bool_array))
print(sum(not_array) <= sum(bool_array))"""


# noinspection PyCompatibility
def optimize_nulls(original: list, new: list):
    """replace null values in address list excluding the """

    optimized_original: list = [original[0]]

    for i in range(1, len(original)):
        if new[i] == '':
            optimized_original.append(original[i])
        else:
            optimized_original.append(new[i])

    return optimized_original


"""start1 = time.perf_counter()
print(optimize_nulls(
    ['Manor House Lane', 10522, 20, 'NY', '', 'Dobbs Ferry'],
    ['MANOR HOUSE LN', 10522, 20, '', '', 'Dobbs Ferry']
))
print(time.perf_counter() - start1)

start2 = time.perf_counter()
print(null_parameters.minimize_nulls(
    ['Manor House Lane', 10522, 20, 'NY', '', 'Dobbs Ferry'],
    ['MANOR HOUSE LN', 10522, 20, '', '', 'Dobbs Ferry'])
)
print(time.perf_counter() - start2)"""

"""a = [[1, 2], [3, 4], [5, 6]]
b = [[1, 2], [1, 3], [2, 3], [3, 4], [4, 6], [5, 6]]
union_a_b = []

a.extend(b)

for pair in a:
    if pair not in union_a_b:
        union_a_b.append(pair)
    else:
        continue

print(union_a_b)"""

"""import requests

url = "https://peekgeo.p.rapidapi.com/reverse_geocode/roof_top"

querystring = {"search_model": "roof_top", "latitude":"43.0920867","longitude":"-77.6439961","radius":"10","reverse_param": False}
print(querystring)
headers = {
    'x-rapidapi-host': "placesGeocoding.p.rapidapi.com",
    'x-rapidapi-key': "2995365d3cmshdc3df09507cf3d7p1252f5jsndbc120dd6f14"
    }

response = requests.request("GET", url, headers=headers, params=querystring)

print(response.text)"""

"""from CythonOperations import radius_arr

start = time.perf_counter()
print(radius_arr.construct_radius_arr(10, 125))
print(time.perf_counter() - start)"""




import sys
import os

rapid_api_token = "bc5b270c17msh058aaac04af67a4p16d4ccjsn75cd807f68f5"
google_api_token = "AIzaSyB94FUFckhQUa-jRpy8Us8HF7AeRQ5KPqU"
estated_token = 'J9aEmitNylIyFD5bQiVLdsZs07qw1t'
mongo_pass = 'Barca2004!'



cpdef bint check_batch(list address_array, int maximum):
    """check if address array length is exceeded"""
    cdef bint counter = False

    if len(address_array) > maximum:
        counter = False
    else:
        counter = True

    return counter


from abc import abstractmethod, ABCMeta
import sys
from time import perf_counter
from places_geocode.app.models import common_models
from elasticsearch.exceptions import ConflictError


# utilize bridge pattern: share the implementation of the logging functionality amongst all input types of input objects
class TrackUsers:
    """track user general class implementation"""

    def __init__(self, implementation):
        # instantialize implementation var
        self._implementation = implementation

    def handle_metrics(self, class_obj, time_now, date, **info_packet):
        """metric base handler function"""
        try:
            # pymongo
            database = class_obj.database_objs[common_models.ObjModel().current_accounts_collection]
        except KeyError:
            # elasticsearch
            database = class_obj.database_objs['connection']

        # create implementation of generalized metric model functionality here
        self._implementation().metric_recorder(class_obj, time_now, date, database, **info_packet)


# noinspection PyCompatibility, PyClassHasNoInit
class UserMetricRecorder(metaclass=ABCMeta):
    """metaclass for user metrics"""

    @abstractmethod
    def metric_recorder(self, class_function, time_now, date, database, **info_packet):
        """
        :param class_function:
        :param time_now:
        :param date:
        :param info_packet:
        """
        pass


# noinspection PyClassHasNoInit
class RadiusMetrics(UserMetricRecorder):
    """implementation for radius metric recorder"""

    def metric_recorder(self, class_function, time_now, date, database, **info_packet):
        """record radius metrics"""
        radius = info_packet['radius']
        token = info_packet['token']

        database.update_one({'token': str(token)},
                            {'$push': {'user_metrics.radius.time': [time_now, date],
                                       'user_metrics.radius.radius': int(radius)}
                             })

        database.update_one({'token': token}, {'$inc': {'user_metrics.radius.hits': 1}})


# noinspection PyClassHasNoInit
class RadiusMetricsElastic(UserMetricRecorder):
    """radius metrics adjusted for elasticsearch"""

    def metric_recorder(self, class_function, time_now, date, database, **info_packet) -> None:
        """record radius metrics"""
        radius = float(info_packet['radius'])
        token = info_packet['token']
        index = common_models.ObjModel().current_accounts_collection

        # create body for user metrics
        body_metrics_time_radius = {
            "script": {
                "inline": "ctx._source.user_metrics.radius.time.add(params.update_time); ctx._source.user_metrics.radius.radius.add(params.radius); "
                          "ctx._source.user_metrics.radius.hits++",
                'params': {
                    'update_time': [time_now, date],
                    'radius': radius
                },
                "lang": "painless"
            },
            "query": {
                "match": {
                    "token": token
                }
            }
        }

        # do updates to database here
        try:
            database.update_by_query(index=index, body=body_metrics_time_radius)
        except ConflictError:
            database.indices.refresh(index=index)
            database.update_by_query(index=index, body=body_metrics_time_radius)


# noinspection PyClassHasNoInit
class ConvexMetrics(UserMetricRecorder):
    """convex metrics implementation"""

    def metric_recorder(self, class_function, time_now, date, database, **info_packet):
        """convex metric record"""
        area = info_packet['area']
        edge_count = info_packet['edge_count']
        token = info_packet['token']

        database.update({'token': token},
                        {'$push': {'user_metrics.convex.time': [time_now,
                                                                date],
                                   'user_metrics.convex.area': area,
                                   'user_metrics.convex.edges_count': edge_count}
                         }
                        )

        database.update({'token': token}, {'$inc': {'user_metrics.convex.hits': 1}})


# noinspection PyClassHasNoInit
class ConvexMetricsElastic(UserMetricRecorder):
    """convex metrics class adjusted for elasticsearch"""

    def metric_recorder(self, class_function, time_now, date, database, **info_packet):
        """metric recorder interior function for convex elasticsearch"""

        # parse for update components
        area = info_packet['area']
        edge_count = info_packet['edge_count']
        token = info_packet['token']

        # collect desired index for elasticsearch update query
        index = common_models.ObjModel().current_accounts_collection

        # create outgoing query bodies
        body_push_metrics = {
            "script": {
                "inline": f"ctx._source.user_metrics.convex.time.add(params.update_time);"
                          f"ctx._source.user_metrics.convex.area.add(params.area);"
                          f"ctx._source.user_metrics.convex.edges_count.add(params.edge_count); "
                          f"ctx._source.user_metrics.convex.hits++",
                "lang": "painless"
            },
            'params': {
                'update_time': [time_now, date],
                'area': area,
                'edges_count': edge_count
            },
            "query": {
                "match": {
                    "token": token
                }
            }
        }

        # compute update queries
        try:
            database.update_by_query(index=index, body=body_push_metrics)
        except ConflictError:
            database.indices.refresh(index=index)
            database.update_by_query(index=index, body=body_push_metrics)


# noinspection PyClassHasNoInit
class DensityMetrics(UserMetricRecorder):
    """density metrics implementation"""

    def metric_recorder(self, class_function, time_now, date, database, **info_packet):
        """record density metrics"""
        unit_in, unit_out, radius, density, custom, token = info_packet['unit_in'], \
                                                            info_packet['unit_out'], \
                                                            info_packet['radius'], \
                                                            info_packet['density'], \
                                                            info_packet['custom'], \
                                                            info_packet['token']

        database.update({'token': token},
                        {'$push': {'user_metrics.density.time': [time_now,
                                                                 date],
                                   'user_metrics.density.radius': radius,
                                   'user_metrics.density.unit_in': unit_in,
                                   'user_metrics.density.unit_out': unit_out,
                                   'user_metrics.density.density': density,
                                   'user_metrics.density.custom': custom
                                   }}
                        )

        database.update({'token': token}, {'$inc': {'user_metrics.density.hits': 1}})


# noinspection PyClassHasNoInit
class DensityMetricsElastic(UserMetricRecorder):
    """density metrics adjusted for elasticsearch"""

    def metric_recorder(self, class_function, time_now, date, database, **info_packet):
        """metric recorder function for elasticsearch"""

        # find the desired update query for elasticsearch
        index = common_models.ObjModel().current_accounts_collection

        # query for update components
        unit_in, unit_out, radius, density, custom, token = info_packet['unit_in'], \
                                                            info_packet['unit_out'], \
                                                            float(info_packet['radius']), \
                                                            info_packet['density'], \
                                                            info_packet['custom'], \
                                                            info_packet['token']

        # create outgoing requests bodies
        body_update_components = {
            "script": {
                "inline": f"ctx._source.user_metrics.density.time.add(params.updated_time); "
                          f"ctx._source.user_metrics.density.radius.add(params.radius); "
                          f"ctx._source.user_metrics.density.unit_in.add(params.unit_in); "
                          f"ctx._source.user_metrics.density.unit_out.add(params.unit_out); "
                          f"ctx._source.user_metrics.density.density.add(params.density); "
                          f"ctx._source.user_metrics.density.custom.add(params.custom); "
                          f"ctx._source.user_metrics.density.hits++",
                'params': {
                    'updated_time': [time_now, date],
                    'radius': radius,
                    'unit_in': unit_in,
                    'unit_out': unit_out,
                    'density': density,
                    'custom': custom
                },
                "lang": "painless"
            },
            "query": {
                "match": {
                    "token": token
                }
            }
        }

        # compute update queries
        try:
            database.update_by_query(index=index, body=body_update_components)
        except ConflictError:
            database.indices.refresh(index=index)
            database.update_by_query(index=index, body=body_update_components)


# noinspection PyClassHasNoInit
class SingleReverseMetrics(UserMetricRecorder):
    """single reverse metrics"""

    def metric_recorder(self, class_function, time_now, date, database, **info_packet):
        """metric recorder for pymongo"""
        coordinates, token = info_packet['coordinates'], info_packet['token']

        database.update({'token': token},
                        {'$push': {'user_metrics.reverse_single.time': [time_now,
                                                                        date],
                                   'user_metrics.reverse_single.coordinates': coordinates}
                         }
                        )

        database.update({'token': token}, {'$inc': {'user_metrics.reverse_single.hits': 1}})


# noinspection PyClassHasNoInit
class SingleReverseMetricsElastic(UserMetricRecorder):
    """single reverse geocoding metrics adjusted for elasticsearch"""

    def metric_recorder(self, class_function, time_now, date, database, **info_packet):
        """metric recorder function for elasticsearch single reverse"""

        # get desired update index for query
        index = common_models.ObjModel().current_accounts_collection

        # get update components
        coordinates, token = info_packet['coordinates'], info_packet['token']

        # create request outgoing bodies
        body_update_components = {
            "script": {
                "inline": f"ctx._source.user_metrics.reverse_single.time.add(params.update_time); "
                          f"ctx._source.user_metrics.reverse_single.coordinates.add(params.coordinates); "
                          f"ctx._source.user_metrics.reverse_single.hits++",
                'params': {
                    'update_time': [time_now, date],
                    'coordinates': coordinates
                },
                "lang": "painless"
            },
            "query": {
                "match": {
                    "token": token
                }
            }
        }

        # compute queries with update one option
        try:
            database.update_by_query(index=index, body=body_update_components)
        except ConflictError:
            database.indices.refresh(index=index)
            database.update_by_query(index=index, body=body_update_components)


# noinspection PyClassHasNoInit
class BatchReverseMetrics(UserMetricRecorder):
    """batch reverse metrics"""

    def metric_recorder(self, class_function, time_now, date, database, **info_packet):
        """batch reverse metric recorder function"""
        batch_size, coordinates, token = info_packet['batch_size'], info_packet['coordinates'], info_packet['token']

        database.update({'token': token},
                        {'$push': {'user_metrics.reverse_batch.time': [time_now,
                                                                       date],
                                   'user_metrics.reverse_batch.coordinates': coordinates,
                                   'user_metrics.reverse_batch.batch_size': batch_size}
                         }
                        )

        database.update({'token': token}, {'$inc': {'user_metrics.reverse_batch.hits': 1}})


# noinspection PyClassHasNoInit
class BatchReverseMetricsElastic(UserMetricRecorder):
    """batch reverse geocoding - elasticsearch adjusted"""

    def metric_recorder(self, class_function, time_now, date, database, **info_packet):
        """metric recorder for elasticsearch batch reverse geocoding"""

        # query common models for desired update index
        index = common_models.ObjModel().current_accounts_collection

        # get update components
        batch_size, coordinates, token = info_packet['batch_size'], info_packet['coordinates'], info_packet['token']

        # create outgoing request bodies
        body_update_components = {
            "script": {
                "inline": f"ctx._source.user_metrics.reverse_batch.time.add(params.update_time); "
                          f"ctx._source.user_metrics.reverse_batch.coordinates.add(params.coordinates); "
                          f"ctx._source.user_metrics.reverse_batch.batch_size.add(params.batch_size); "
                          f"ctx._source.user_metrics.reverse_batch.hits++",
                'params': {
                    'update_time': [time_now, date],
                    'coordinates': coordinates,
                    'batch_size': batch_size
                },
                "lang": "painless"
            },
            "query": {
                "match": {
                    "token": token
                }
            }
        }

        # compute update one queries
        try:
            database.update_by_query(index=index, body=body_update_components)
        except ConflictError:
            database.indices.refresh(index=index)
            database.update_by_query(index=index, body=body_update_components)


# noinspection PyClassHasNoInit
class SingleForwardMetrics(UserMetricRecorder):
    """single forward geocoding metrics here"""

    def metric_recorder(self, class_function, time_now, date, database, **info_packet):
        """metric recorder for pymongo"""
        address, parsed_counter, full_counter, token = info_packet['address'], \
                                                       info_packet['parsed'], \
                                                       info_packet['full'], \
                                                       info_packet['token']

        database.update({'token': token},
                        {'$push': {'user_metrics.geocode_single.time': [time_now,
                                                                        date],
                                   'user_metrics.geocode_single.address': address}
                         }
                        )

        if full_counter == 1:
            database.update({'token': token},
                            {'$inc': {'user_metrics.geocode_single.full_address_count': 1,
                                      'user_metrics.geocode_single.hits': 1}})
        elif parsed_counter == 1:
            database.update({'token': token},
                            {'$inc': {
                                'user_metrics.geocode_single.parsed_address_count': 1,
                                'user_metrics.geocode_single.hits': 1}})


# noinspection PyClassHasNoInit
class SingleForwardMetricsElastic(UserMetricRecorder):
    """single forward geocoding adjusted for elasticsearch"""

    def metric_recorder(self, class_function, time_now, date, database, **info_packet):
        """metric recorder for single forward geocoding - elasticsearch engine"""

        # get desired update index from common models
        index = common_models.ObjModel().current_accounts_collection

        # query for update components
        address, parsed_counter, full_counter, token = info_packet['address'], \
                                                       info_packet['parsed'], \
                                                       info_packet['full'], \
                                                       info_packet['token']

        # create outgoing request bodies for counter variables
        if full_counter == 1:
            body_update_components = {
                "script": {
                    "inline": f"ctx._source.user_metrics.geocode_single.time.add(params.update_time); "
                              f"ctx._source.user_metrics.geocode_single.hits++; "
                              f"ctx._source.user_metrics.geocode_single.full_address_count++",
                    'params': {
                        'update_time': [time_now, date],
                    },
                    "lang": "painless"
                },
                "query": {
                    "match": {
                        "token": token
                    }
                }
            }
        elif parsed_counter == 1:
            # create outgoing request bodies for component update variables
            body_update_components = {
                "script": {
                    "inline": f"ctx._source.user_metrics.geocode_single.time.add(params.update_time); "
                              f"ctx._source.user_metrics.geocode_single.hits++; "
                              f"ctx._source.user_metrics.geocode_single.parsed_address_count++",
                    'params': {
                        'update_time': [time_now, date],
                    },
                    "lang": "painless"
                },
                "query": {
                    "match": {
                        "token": token
                    }
                }
            }
        else:
            body_update_components = {}

        # compute update queries in elasticsearch
        try:
            database.update_by_query(index=index, body=body_update_components)
        except ConflictError:
            database.indices.refresh(index='apiusers')
            database.update_by_query(index=index, body=body_update_components)


# noinspection PyClassHasNoInit
class BatchForwardMetrics(UserMetricRecorder):
    """batch forward geocoding metrics here"""

    def metric_recorder(self, class_function, time_now, date, database, **info_packet):
        """metric recorder for pymongo"""
        addresses, batch_size, token = info_packet['addresses'], info_packet['batch_size'], info_packet['token']

        database.update({'token': token},
                        {'$push': {'user_metrics.geocode_batch.time': [time_now,
                                                                       date],
                                   'user_metrics.geocode_batch.address': addresses,
                                   'user_metrics.geocode_batch.batch_size': batch_size}
                         }
                        )

        database.update({'token': token}, {'$inc': {'user_metrics.geocode_batch.hits': 1}})


# noinspection PyClassHasNoInit
class BatchForwardMetricsElastic(UserMetricRecorder):
    """batch forward geocoding metrics for elasticsearch"""

    def metric_recorder(self, class_function, time_now, date, database, **info_packet):
        """metric recorder for elastic search batch geocoding"""

        # find desired update query index
        index = common_models.ObjModel().current_accounts_collection

        # query for update components
        addresses, batch_size, token = info_packet['addresses'], str(info_packet['batch_size']), info_packet['token']

        # create outgoing request bodies
        body_update_components = {
            "script": {
                "inline": f"ctx._source.user_metrics.geocode_batch.time.add(params.update_time); "
                          f"ctx._source.user_metrics.geocode_batch.batch_size.add(params.batch_size); "
                          f"ctx._source.user_metrics.geocode_batch.hits++",
                'params': {
                    'update_time': [time_now, date],
                    'batch_size': batch_size
                },
                "lang": "painless"
            },
            "query": {
                "match": {
                    "token": token
                }
            }
        }

        # compute update queries for elasticsearch
        try:
            database.update_by_query(index=index, body=body_update_components)
        except ConflictError:
            database.indices.refresh(index=index)
            database.update_by_query(index=index, body=body_update_components)


# noinspection PyClassHasNoInit
class SpreadsheetReverseMetrics(UserMetricRecorder):
    """spreadsheet reverse metric handler"""

    def metric_recorder(self, class_function, time_now, date, **info_packet):
        batch_size, token = info_packet['batch_size'], info_packet['token']

        class_function.api_accounts.update({'token': token}, {'$inc': {'user_metrics.reverse_large_scale.hits': 1}})

        class_function.api_accounts.update({'token': token},
                                           {'$push': {'user_metrics.reverse_large_scale.time': [time_now,
                                                                                                date],
                                                      'user_metrics.reverse_large_scale.batch_size': batch_size}
                                            }
                                           )


# noinspection PyClassHasNoInit
class SpreadsheetForwardMetrics(UserMetricRecorder):
    """implementation of spreadsheet forward geocoding pattern"""

    def metric_recorder(self, class_function, time_now, date, **info_packet):
        batch_size, token = info_packet['batch_size'], info_packet['token']

        class_function.api_accounts.update({'token': token}, {'$inc': {'user_metrics.geocode_large_scaled.hits': 1}})

        class_function.api_accounts.update({'token': token},
                                           {'$push': {'user_metrics.geocode_large_scaled.time': [time_now,
                                                                                                 date],
                                                      'user_metrics.geocode_large_scaled.batch_size': batch_size}
                                            }
                                           )



import csv

cpdef write_csv(int filename, int mode, list results):
    cdef const char *fieldnames[3]
    fieldnames[:3] = ['address', 'latitude', 'longitude']
    cdef list result = []
    cdef list data = []
    cdef dict d = {}
    cdef list coordinates = []

    with open(filename, mode) as csvfile:
        csvfile.truncate()
        writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
        writer.writeheader()

        for result in results:
            data = result[0]
            for d in data:
                coordinates = d['coordinates']
                writer.writerow({'address': d['formatted_address'], 'latitude': d['coordinates'][0],
                                 'longitude': coordinates[1]})


"""Cython function for checking incorrect lat, long scenarios"""

cpdef bint check_lat_long(list latitudes, list longitudes, int maximum):
    """check if lat long are the same length"""

    cdef bint counter = False
    cdef int coord_length = len(latitudes)

    if len(latitudes) == len(longitudes):
        if coord_length <= maximum:
            counter = True

    return counter






# noinspection PyTypeChecker
cpdef list coords_conv(list coordinates):
    """convert coordinates to lats and longs"""

    # create containers for lats and longs
    cdef list lats = []
    cdef list longs = []

    # separate the latitudes and longitudes
    [(lats.append(i[0]), longs.append(i[1])) for i in coordinates]

    return [lats, longs]


# noinspection PyTypeChecker
cpdef dict executor_read(list results):
    cdef dict response_packet = {'response_packet': [], 'supplied_packet': []}
    cdef dict result = {}

    for result in results:
        response_packet['response_packet'].extend(result['response_packet'])
        response_packet['supplied_packet'].extend(result['supplied_packet'])

    return response_packet





import pydantic
from typing import List, Optional, Union, Any
import sys


# noinspection PyCompatibility
class ObjModel(pydantic.BaseModel):
    """keep track of maxSize parameter of client initialization"""
    maxSize: Union[int, None]
    current_geographic_collection: Optional[str] = 'proplocations'
    current_accounts_collection: Optional[str] = 'apiusers'
    token_choices: Optional[List[str]] = list("0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ")


poi_not_found_check = lambda formatted_address: formatted_address == ''


# noinspection PyCompatibility
class CommonUtilityModel(pydantic.BaseModel):
    """common model for all overarching tasks"""
    earth_radius: Optional[float] = 6378.1 * 1000
    generator_threshold: Optional[int] = 250
    states: Optional[List[str]] = ["AL", "AK", "AZ", "AR", "CA", "CO", "CT", "DC", "DE", "FL", "GA",
                                   "HI", "ID", "IL", "IN", "IA", "KS", "KY", "LA", "ME", "MD",
                                   "MA", "MI", "MN", "MS", "MO", "MT", "NE", "NV", "NH", "NJ",
                                   "NM", "NY", "NC", "ND", "OH", "OK", "OR", "PA", "RI", "SC",
                                   "SD", "TN", "TX", "UT", "VT", "VA", "WA", "WV", "WI", "WY"]


# noinspection PyCompatibility
class PointsModel(pydantic.BaseModel):
    """points model here"""
    coordinates: List[List[Union[float, int]]]

    @pydantic.validator("coordinates")
    @classmethod
    def coordinate_val(cls, arr) -> List:
        if not isinstance(arr, list):
            raise Exception("Invalid Coordinate Array Inputted")
        else:
            return arr


def parse_coordinates_elastic(document: dict) -> List[Union[float, Any]]:
    """parse the elastic coordinates from elastic DB"""

    # get the location db object
    location_key: str = document['location']

    # remove whitespaces
    location_key = location_key.replace(" ", "")

    # capture latitude and longitude
    latitude, longitude = location_key.split(",")

    # type conversion for lat/long pair
    try:
        type_conv_latitude = float(latitude)
    except TypeError:
        type_conv_latitude = latitude

    try:
        type_conv_longitude = float(longitude)
    except TypeError:
        type_conv_longitude = longitude

    return [type_conv_latitude, type_conv_longitude]


def parse_coordinates_pymongo(document: dict) -> List[Union[int, float, None]]:
    """parse the pymongo document coordinates"""

    return document['location']['coordinates']


from difflib import SequenceMatcher
from bson import BSON

# noinspection PyTypeChecker,PyArgumentEqualDefault
cpdef list find_similar_components(str street, str region, str city, result):

    """find the most similar address components using the extension to the TFIDF-ML algorithm"""

    cdef float current_max_street = 0.0
    cdef float current_max_region = 0.0
    cdef float current_max_city = 0.0
    cdef dict current_max_result = {'street': None, 'city': None, 'region': None}
    cdef dict street_comparison = {}
    cdef float similarity_street = 0.0
    cdef float similarity_region = 0.0
    cdef float similarity_city = 0.0

    cdef dict property_components = {}

    for document in result:

        document = document['properties']

        # find similarities
        similarity_street = float(SequenceMatcher(None, street.lower(),
                                           document['street'].lower()).ratio())

        similarity_city = float(SequenceMatcher(None, street.lower(),
                                           document['city'].lower()).ratio())

        similarity_region = float(SequenceMatcher(None, street.lower(),
                                                document['region'].lower()).ratio())

        if similarity_street > current_max_street:
            current_max_street = similarity_street
            current_max_result['street'] = document

        if similarity_city > current_max_city:
            current_max_city = similarity_city
            current_max_result['city'] = document

        if similarity_region > current_max_region:
            current_max_region = similarity_region
            current_max_result['region'] = document

    return [
        [round(current_max_street, 2), current_max_result['street'], current_max_result['street']],
        [round(current_max_city, 2), current_max_result['city'], current_max_result['city']],
        [round(current_max_region, 2), current_max_result['region'], current_max_result['region']]
    ]



"""finite state machine (event-based) for unit conversion functionality on density endpoint"""

from state_machine import State, Event, acts_as_state_machine, after, InvalidStateTransition
from typing import Union, Callable, Optional, Any
from pydantic import BaseModel


# noinspection PyCompatibility
class StatesModel(BaseModel):
    """states and events base model"""

    # define the necessary states
    arbitrary_unit: Optional[Any] = State(initial=True)
    kilometer: Optional[Any] = State()
    mile: Optional[Any] = State()
    yard: Optional[Any] = State()
    foot: Optional[Any] = State()
    meter: Optional[Any] = State()

    class Config:
        """config class for arbitrary types"""
        arbitrary_types_allowed = True


# noinspection PyCompatibility
class EventsModelInput(BaseModel):
    """create events base model"""

    # create the inner states model and arbitrary unit type
    states = StatesModel()
    arbitrary_unit = states.arbitrary_unit

    # define processes/events
    to_ft: Optional[Any] = Event(from_states=arbitrary_unit, to_state=states.foot)
    to_mi: Optional[Any] = Event(from_states=arbitrary_unit, to_state=states.mile)
    to_m: Optional[Any] = Event(from_states=arbitrary_unit, to_state=states.meter)
    to_yd: Optional[Any] = Event(from_states=arbitrary_unit, to_state=states.yard)
    to_km: Optional[Any] = Event(from_states=arbitrary_unit, to_state=states.kilometer)

    class Config:
        """config class for arbitrary types"""
        arbitrary_types_allowed = True


# noinspection PyCompatibility
class EventsModelOutput(BaseModel):
    """output events defined in base model"""
    # create the inner states model and arbitrary unit type
    states = StatesModel()
    arbitrary_unit = states.arbitrary_unit

    # define events
    to_ft_output: Optional[Any] = Event(from_states=arbitrary_unit, to_state=states.foot)
    to_mi_output: Optional[Any] = Event(from_states=arbitrary_unit, to_state=states.mile)
    to_m_output: Optional[Any] = Event(from_states=arbitrary_unit, to_state=states.meter)
    to_yd_output: Optional[Any] = Event(from_states=arbitrary_unit, to_state=states.yard)
    to_km_output: Optional[Any] = Event(from_states=arbitrary_unit, to_state=states.kilometer)

    class Config:
        """config class for arbitrary types"""
        arbitrary_types_allowed = True


# noinspection PyCompatibility
@acts_as_state_machine
class UnitConversionsToFeet:

    """track the unit conversions"""

    # define the necessary states
    arbitrary_unit: Optional[Any] = State(initial=True)
    kilometer: Optional[Any] = State()
    mile: Optional[Any] = State()
    yard: Optional[Any] = State()
    foot: Optional[Any] = State()
    meter: Optional[Any] = State()

    # define processes/events
    to_ft: Optional[Any] = Event(from_states=arbitrary_unit, to_state=foot)
    to_mi: Optional[Any] = Event(from_states=arbitrary_unit, to_state=foot)
    to_m: Optional[Any] = Event(from_states=arbitrary_unit, to_state=foot)
    to_yd: Optional[Any] = Event(from_states=arbitrary_unit, to_state=foot)
    to_km: Optional[Any] = Event(from_states=arbitrary_unit, to_state=foot)

    # define output events
    to_m_output: Optional[Any] = Event(from_states=arbitrary_unit, to_state=meter)
    to_km_output: Optional[Any] = Event(from_states=arbitrary_unit, to_state=kilometer)
    to_yd_output: Optional[Any] = Event(from_states=arbitrary_unit, to_state=yard)
    to_ft_output: Optional[Any] = Event(from_states=arbitrary_unit, to_state=foot)
    to_mi_output: Optional[Any] = Event(from_states=arbitrary_unit, to_state=mile)

    def __init__(self, radius: Union[float, int], *conversion_algorithm):
        self.radius = radius
        self.unit_result: Union[int, float, None] = None
        self.conv_algorithm_outer: Union[None, Callable] = None
        self.conv_algorithm_inner: Union[None, Callable] = None

    @staticmethod
    def call_event(event: Callable) -> None:
        """call a class event"""
        event()

    @after('to_ft')
    def feet_conversion(self):
        """convert radius to feet"""
        self.unit_result = self.radius

    @after('to_m')
    def meter_conversion(self):
        """convert radius to meters"""
        self.unit_result = self.conv_algorithm_outer(self.radius)

    @after('to_yd')
    def yard_conversion(self):
        """convert radius to yards"""
        self.unit_result = self.radius * 3

    @after('to_mi')
    def mile_conversion(self):
        """convert radius to miles"""
        self.unit_result = self.conv_algorithm_outer(self.radius)

    @after('to_km')
    def km_conversion(self):
        """convert radius to kilometers"""
        self.unit_result = self.conv_algorithm_outer(self.conv_algorithm_inner(self.radius))

    @after('to_ft_output')
    def to_feet_conversion(self):
        """radius to feet for output"""
        self.unit_result = self.radius

    @after('to_m_output')
    def to_meter_conversion(self):
        """radius to meter for output"""
        self.unit_result = self.conv_algorithm_outer(self.radius)

    @after('to_mi_output')
    def to_mile_conversion(self):
        """radius to mile unit for output"""
        self.unit_result = self.conv_algorithm_outer(self.radius)

    @after('to_yd_output')
    def to_yard_conversion(self):
        """to yard conversion for radius output"""
        self.unit_result = self.conv_algorithm_outer(self.radius)

    @after('to_km_output')
    def to_kilometer_conversion(self):
        """radius to kilometers for output"""
        self.unit_result = self.conv_algorithm_inner(self.conv_algorithm_outer(self.radius))


# noinspection PyTypeHints,PyCompatibility
def state_machine_conversion_controller(conv_option, conv_class, *conversion_algorithm):
    """control the finite-state machine conversion process"""

    # process the *args conversion algorithms input
    try:
        conv_class.conv_algorithm_outer: Union[None, Callable] = conversion_algorithm[0]
    except IndexError:
        conv_class.conv_algorithm_outer: Union[None, Callable] = None

    try:
        conv_class.conv_algorithm_inner: Union[None, Callable] = conversion_algorithm[1]
    except IndexError:
        conv_class.conv_algorithm_inner: Union[None, Callable] = None

    # run the event
    try:
        conv_class.call_event(conv_option)
        return conv_class.unit_result
    except InvalidStateTransition:
        raise BufferError










from difflib import SequenceMatcher


"""TFIDF algorithm implementation Cython"""
# noinspection PyTypeChecker,PyArgumentEqualDefault
cpdef list find_max_similarity(str street_query, list result):

    """
    :param street_query: street to be compared to
    :param result: general query output 
    :return: maximum accuracy achieved, JSON of result, street: str
    """

    cdef float current_max = 0.0
    cdef dict current_max_result = {}
    cdef dict street_comparison = {}
    cdef dict properties = {}
    cdef float similarity = 0.0

    for street_comparison in result:
        properties = street_comparison['_source']['properties']
        similarity = float(SequenceMatcher(None, street_query.lower(),
                                     properties['street'].lower()).ratio())

        if similarity > current_max:
            current_max = similarity
            current_max_result = street_comparison
        else:
            continue

    return [round(current_max, 2), current_max_result['_source'], current_max_result['_source']['properties']['street']]


import sys
import pydantic
from typing import Union, Optional
sys.path.insert(1, '/app')


# noinspection PyCompatibility
class ConvexBaseData(pydantic.BaseModel):
    area_limit: Optional[Union[float, int]] = 250
    edge_limit: Optional[int] = 20



"""test the PeekGeo package"""
"""from placesGeocoding.peek_geo import peekgeo_api as pg
from time import perf_counter

# create geocoder object
geocoder = pg.PeekGeo('eyJhbGci')

start = perf_counter()
# compute query output
print(geocoder.batch_forward(
    ['20 Manor House Lane, Dobbs Ferry, NY, 10522',
     '18 Manor House Lane, Dobbs Ferry, NY, 10522',
     '16 Manor House Lane, Dobbs Ferry, NY, 10522',
     '29 Manor House Lane, Dobbs Ferry, NY, 10522',
     '31 Manor House Lane, Dobbs Ferry, NY, 10522',
     '35 Manor House Lane, Dobbs Ferry, NY, 10522']
))
print(perf_counter() - start)"""




"""PeekGeo API endpoint program"""

# from fastapi.responses import StreamingResponse
from fastapi import HTTPException, Depends, Query, UploadFile, File
from dependency_injector import containers, providers
from dependency_injector.wiring import inject, Provide, Provider

# dataclass modules
from typing import Optional, List, Union, Callable, Tuple, Any, Dict
from pydantic import BaseModel

# runtime utilities
import os
import sys

# noinspection PyCompatibility
from asyncio import create_task, run

# time modules
from datetime import datetime, date
from time import perf_counter
from ast import literal_eval

# elasticsearch imports
from elasticsearch.exceptions import ConflictError, RequestError

# data science/helper-methods imports
import json
import importlib

from places_geocode.app.geocodingBackend import GeneralGeocodingManager, GeneralReverseManager, DensityRadius, \
    Convex, RunModel

# models import
from places_geocode.app.models import frontend_models, density_models, common_models, user_metrics_models as USM
from places_geocode.app.logger import log_error

path = os.path.dirname(os.path.realpath(__file__))

# import cython files
from ctypes import *
try:
    create_convex_array = cdll.LoadLibrary("create_convex_array.cpython-38-darwin.so".format(path))
    check_dups = cdll.LoadLibrary("check_dups.cpython-38-darwin.so".format(path))
    zip_coordinate = cdll.LoadLibrary("zip_coordinate.cpython-38-darwin.so".format(path))
    check_lat_long_length = cdll.LoadLibrary("check_lat_long_length.cpython-38-darwin.so".format(path))
    check_address_length = cdll.LoadLibrary("check_address_length.cpython-38-darwin.so".format(path))
    zip_coordinates_elastic = cdll.LoadLibrary("zip_coordinates_elastic.cpython-38-darwin.so".format(path))
    create_convex_elastic = cdll.LoadLibrary("create_convex_elastic.cpython-38-darwin.so".format(path))
    coordinate_conv_package = cdll.LoadLibrary("coordinate_conv_package.cpython-38-darwin.so".format(path))
except:
    from places_geocode.app.dprk_copy import create_convex_array, check_dups, zip_coordinate, check_lat_long_length, \
        check_address_length, zip_coordinates_elastic, create_convex_elastic, coordinate_conv_package


# noinspection PyCompatibility
class GlobalMetricManager(BaseModel):
    """global metric manager for handling metrics of each endpoint"""
    load_radius: Optional[str] = 'radius'
    convex: Optional[str] = 'convex'
    density: Optional[str] = 'density'
    reverse_single: Optional[str] = 'reverse_single'
    reverse_batch: Optional[str] = 'reverse_batch'
    single_forward: Optional[str] = 'geocode_single'
    batch_forward: Optional[str] = 'geocode_batch'
    api_call_key: Optional[str] = 'api_calls'


class SingletonType(type):
    """singleton class for singular base url instantiation"""
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(SingletonType, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


# noinspection PyCompatibility,PyTypeChecker
class PeekGeoUtilities(metaclass=SingletonType):
    """handle all backend object creation by creating only one backend object only when needed"""

    def __init__(self, backend_method: Any):
        # create settings models
        self.global_model: GlobalDataModel = GlobalDataModel()
        self.runner_model: RunModel = RunModel()

        # create backend objects for storing databases and functionality
        self.backend_utility = backend_method(self.global_model.database, self.global_model.collections)

        # mappings for database options
        self.coordinate_mapping: dict = {'pymongo': zip_coordinate,
                                         'elastic': zip_coordinates_elastic}
        self.convex_mapping: dict = {'pymongo': create_convex_array, 'elastic': create_convex_elastic}
        self.metric_handler_mapping: dict = {
            'pymongo': {
                'radius': USM.RadiusMetrics,
                'convex': USM.ConvexMetrics,
                'density': USM.DensityMetrics,
                'reverse_single': USM.SingleReverseMetrics,
                'reverse_batch': USM.BatchReverseMetrics,
                'forward_single': USM.SingleForwardMetrics,
                'forward_batch': USM.BatchForwardMetrics
            },
            'elastic': {
                'radius': USM.RadiusMetricsElastic,
                'convex': USM.ConvexMetricsElastic,
                'density': USM.DensityMetricsElastic,
                'reverse_single': USM.SingleReverseMetricsElastic,
                'reverse_batch': USM.BatchReverseMetricsElastic,
                'forward_single': USM.SingleForwardMetricsElastic,
                'forward_batch': USM.BatchForwardMetricsElastic
            }
        }
        self.custom_density_mapping: dict = {'pymongo': {'state': density_models.StateDensityStrategy(),
                                                         'city': density_models.CityDensityStrategy(),
                                                         'postcode': density_models.PostcodeDensityStrategy()},
                                             'elastic': {'state': density_models.StateDensityElastic(),
                                                         'city': density_models.CityDensityElastic(),
                                                         'postcode': density_models.PostcodeDensityElastic()}
                                             }

    @staticmethod
    async def background_logging(name: str, filename: str, logger_level: str, message: str) -> None:
        """conduct background logging activity"""
        try:
            log_error(name, filename, logger_level, message)
        except FileExistsError or FileNotFoundError:
            raise BrokenPipeError

    @staticmethod
    async def get_datetime_objs() -> Dict:
        """create the datetime objects"""
        date_now = str(date.today())
        time_now = str(datetime.now())
        official_timestamp = datetime.timestamp(datetime.now())
        return dict(date=date_now, time=time_now, official_timestamp=official_timestamp)

    @staticmethod
    def handle_coordinates(coordinates: list, reverse_param: bool):
        """handle reversals and such"""
        if reverse_param:
            return coordinates[::-1]
        else:
            return coordinates

    def _clean_file_batch_reverse(self, coordinate_file: Any):

        """clean the batch reverse geocoding file for elasticsearch and pymongo differentiation"""

        coord_data = json.load(coordinate_file.file)['coordinates']
        if self.runner_model.engine == 'pymongo':
            return coord_data
        else:
            stringified_coord_data = []
            for coordinate in coord_data:
                coordinate = f'{coordinate[0]}, {coordinate[1]}'
                stringified_coord_data.append(coordinate)
            return stringified_coord_data

    @staticmethod
    def handle_batch_reverse_coords(latitudes: list, longitudes: list,
                                    reverse_param: bool,
                                    coordinate_file: UploadFile = File(None)) -> Any:

        """handle the spreadsheet and batch coordinate reverse geocoding logic"""
        """FILE INPUT HAS PRIORITY"""

        if coordinate_file is not None:
            return json.load(coordinate_file.file)['coordinates']
        else:
            if latitudes is None and longitudes is None:
                raise HTTPException(
                    status_code=505,
                    detail='No Input Detected',
                    headers={"Error_Description": "Check documentation for input requirements"}
                )
            else:
                # check the length of the coordinates to not exceed batch limit
                if check_lat_long_length.check_lat_long(latitudes, longitudes, GlobalDataModel().batch_endpoint_max):
                    # return based on the database engine being run
                    return zip_coordinate.coordinate_zipped(latitudes, longitudes, reverse_param)
                else:
                    raise HTTPException(
                        status_code=505,
                        detail='Array Mapping Error Encountered',
                        headers={
                            "Error_Description": "Latitudes List length must match Longitudes List length. Maximum batch is {}".format(
                                GlobalDataModel().batch_endpoint_max)}
                    )

    @staticmethod
    def handle_convex_coordinates(latitudes, longitudes, reverse_param) -> list:
        """handle the coordinates of the convex endpoint"""

        return create_convex_array.convex_array(latitudes, longitudes, reverse_param)

    @staticmethod
    def handle_batch_reverse_addresses(addresses: List[str], address_file: UploadFile = File(None)):
        """handle the batch address calls, must have addresses as header"""
        if address_file is not None:
            # noinspection PyTypeChecker
            return json.loads(address_file.file)['addresses']
        else:
            if addresses is None:
                raise HTTPException(
                    status_code=505,
                    detail='No Address Input Detected',
                    headers={"Error_Description": "Check documentation for address input requirements"}
                )
            else:
                if check_address_length.check_batch(addresses, GlobalDataModel().batch_endpoint_max):
                    return addresses
                else:
                    raise HTTPException(
                        status_code=505,
                        detail='Batch Size Exceeded',
                        headers={
                            "Error_Description": "Maximum for batch forward geocoding was exceeded. Maximum batch is {}".format(
                                GlobalDataModel().batch_endpoint_max)}
                    )

    def close_client(self) -> None:
        """close the mongo client"""
        try:
            self.backend_utility.database_objs['client'].close()
        except KeyError:
            pass

    def background_api_call_handler(self, token: str, delta_call: int, original_calls: int, method: str) -> None:
        """all background api calls handled for website version"""
        try:
            # create the call update delta
            updated_call_count = original_calls - delta_call
            if updated_call_count < 0:
                updated_call_count = 0

            # query for updating one document from user accounts db
            try:
                # pymongo option
                self.backend_utility.database_objs[common_models.ObjModel().current_accounts_collection].update(
                    {'token': token}, {'$set': {f'{method}': updated_call_count}})
            except KeyError:
                # elasticsearch option, create query request body
                body_calls_update = {
                    "script": {
                        "inline": f"ctx._source.{method}={updated_call_count}",
                        "lang": "painless"
                    },
                    "query": {
                        "match": {
                            "token": token
                        }
                    }
                }

                # compute query of update_one document
                try:
                    self.backend_utility.database_objs['connection'].update_by_query(
                        index=common_models.ObjModel().current_accounts_collection,
                        body=body_calls_update)
                except ConflictError:
                    self.backend_utility.database_objs['connection'].indices.refresh(index=common_models.ObjModel().current_accounts_collection)
                    self.backend_utility.database_objs['connection'].update_by_query(
                        index=common_models.ObjModel().current_accounts_collection,
                        body=body_calls_update)

        except TypeError or ValueError:
            raise Exception('Background API call handler encountered error state')

    def get_user(self, username: str):
        """get the user by username, used for security when logging in with username"""
        try:
            user = list(self.backend_utility.database_objs[common_models.ObjModel().current_accounts_collection].find(
                {'username': username}))[0]
        except KeyError:
            # create outgoing body request
            body = {
                "query": {
                    "bool": {
                        "should": [
                            {"term": {"username": username}}
                        ]
                    }
                }
            }

            # conduct elasticsearch query
            try:
                user = self.backend_utility.database_objs['connection'].search(
                    index=common_models.ObjModel().current_accounts_collection, body=body)['hits']['hits']
            except KeyError or IndexError:
                user = None

        return user

    def get_call_limit(self, token: str, method: str) -> Any:
        """get the user api call maximum"""
        try:
            try: # mongodb version
                query = list(self.backend_utility.database_objs[
                    common_models.ObjModel().current_accounts_collection
                ].find({'token': token}))[0][f'{method}']
            # create outgoing request body
            except KeyError:
                body = {
                    "query": {
                        "match": {
                            "token": token
                        }
                    }
                }

                # conduct elasticsearch query
                query = self.backend_utility.database_objs['connection'].search(
                        index=common_models.ObjModel().current_accounts_collection, body=body)['hits']['hits'][0]['_source'][f'{method}']

            return query

        except (IndexError, TypeError, RequestError):
            raise HTTPException(
                status_code=500,
                detail="Invalid API Token Encountered",
                headers={'Error_Description': "Reset token or validate current token"}
            )

    def verify_token(self, token: str) -> Any:
        """verify the access token to API endpoint"""

        try:
            query = len(
                list(self.backend_utility.database_objs[common_models.ObjModel().current_accounts_collection].find(
                    {'token': token})))
        except KeyError:
            # create request body
            body = {
                "query": {
                    "bool": {
                        "should": [
                            {"term": {"token": token}}
                        ]
                    }
                }
            }

            # conduct elastic search with exception for key errors upon receiving incoming request
            try:
                query = len(self.backend_utility.database_objs['connection'].search(
                    index=common_models.ObjModel().current_accounts_collection,
                    body=body)['hits']['hits'])
            except KeyError or IndexError or TypeError:
                query = 0

        return query == 0

    async def handle_user_metrics(self, metric_method: str, **kwargs):
        """handle user metrics method for recording behavior and progress of each person in account data packet"""
        date_results = create_task(self.get_datetime_objs())
        date_results = await date_results
        time_now, date_now, official_timestamp = date_results.values()

        # use the bridge pattern for the handler objects in the user_models module
        try:
            metric_handler = USM.TrackUsers(self.metric_handler_mapping[self.runner_model.engine][metric_method])
            metric_handler.handle_metrics(self.backend_utility, time_now, date_now, **kwargs)
        except KeyError:
            raise Exception("User Metrics failed")

    def account_data(self, token_query: dict, method: str) -> Tuple[int, int, float, datetime]:
        """get account data for specified token document"""

        # get the account date, pay plan from user document
        account_date, plan = str(token_query['date']), token_query['plan']

        # get account datetime object, create current object
        account_date, current_date = datetime.fromisoformat(account_date.replace("Z", "")), str(datetime.now()).replace(" ", "T")

        # compute time difference in minutes
        delta_time = datetime.fromisoformat(current_date) - account_date
        minutes = divmod(delta_time.total_seconds(), 60)[0]

        # find api calls and spreadsheet calls
        api_calls, spreadsheet_calls = self.global_model.call_pay_plans[plan]

        if minutes >= 1440:
            return api_calls, spreadsheet_calls, minutes, current_date
        else:
            if method == 'single':
                return token_query['api_calls'] - 1, token_query['spreadsheet_calls'], minutes, current_date
            else:
                return token_query['api_calls'], token_query['spreadsheet_calls'] - 1, minutes, current_date

    async def regulate_api_calls(self, token: str, method: str) -> None:
        """regulate and renew API calls"""

        # find the account using the user's oauth2 token
        try:
            # pymongo version
            token_query = list(
                self.backend_utility.database_objs[common_models.ObjModel().current_accounts_collection].find(
                    {'token': token}))[0]
        except KeyError:
            # elasticsearch version

            # create outgoing request body
            body = {
                "query": {
                    "match": {
                        'token': token
                    }
                }
            }

            # conduct elasticsearch query
            try:
                token_query = self.backend_utility.database_objs['connection'].search(
                    index=common_models.ObjModel().current_accounts_collection,
                    body=body)['hits']['hits'][0]['_source']
            except (KeyError, IndexError):
                token_query = {}

        try:
            # get account information based on token query
            api_calls, spreadsheet_calls, minutes, current_date = self.account_data(token_query, method)

            # condition for call regeneration (once every day, one minute before)
            try:
                # pymongo update one document
                self.backend_utility.database_objs[common_models.ObjModel().current_accounts_collection].update_one(
                    {'token': token}, {'$set': {'api_calls': api_calls, 'spreadsheet_calls': spreadsheet_calls,
                                                'date': current_date}})

            except KeyError:
                # create request body for elasticsearch update query
                body_api_calls = {
                    "script": {
                        "inline": f"ctx._source.api_calls={api_calls}; "
                                  f"ctx._source.spreadsheet_calls={spreadsheet_calls}; "
                                  f"ctx._source.date=ctx._source.date=new SimpleDateFormat('yyyy-MM-dd').parse('{current_date}')",
                        "lang": "painless"
                    },
                    "query": {
                        "match": {
                            "token": token
                        }
                    }
                }

                # compute update queries
                self.backend_utility.database_objs['connection'].update_by_query(
                    index=common_models.ObjModel().current_accounts_collection,
                    body=body_api_calls)

        except (IndexError, KeyError):
            raise Exception("Unidentified User")


############################# SETTINGS MODEL ###############################


# noinspection PyCompatibility
class GlobalDataModel(BaseModel):
    """global data class for common variables across all endpoints"""

    database: Optional[str] = 'PeekDB'
    collections: Optional[Tuple] = ("proplocations", "apiusers")
    convex_edge_limit: Optional[int] = 20
    convex_area_limit: Optional[int] = 250
    density_radius_limit: Optional[dict] = {'ft': 9843, 'm': 3000.146, 'mi': 1.864205, 'yd': 3281, 'km': 3.000146}
    density_strategy_table: Optional[dict] = {'state': density_models.StateDensityStrategy(),
                                              'city': density_models.CityDensityStrategy(),
                                              'postcode': density_models.PostcodeDensityStrategy()}
    batch_endpoint_max: Optional[int] = 125
    max_load_radius: Optional[int] = 2000
    token_length: Optional[int] = 12
    call_pay_plans: Optional[Dict[Any, Tuple[int, int]]] = {0: (100, 0),
                                                                        20: (13000, 20),
                                                                        60: (100000, 50),
                                                                        125: (350000, 200)}


################################## Error Response BaseClasses #####################################


# noinspection PyCompatibility
class NullResponses(BaseModel):
    """null responses for all api calls"""

    exhausted_calls: Optional[Dict] = {
        'meta': {'build': 'NO_CALLS', 'status': 404, 'matching': 0,
                 'timestamps': {'date': str(date.today()),
                                'time': str(datetime.now()),
                                'official_timestamp': datetime.timestamp(datetime.now())},
                 'geocoded_data': {}},
        'supplied_data': {},
        'source': '©OpenStreetMap contributors'
    }

    internal_error: Optional[Dict] = {'meta': {'build': 'INTERNAL_ERR', 'status': 500, 'matching': 0,
                                               'timestamps': {'date': str(date.today()),
                                                              'time': str(datetime.now()),
                                                              'official_timestamp': datetime.timestamp(
                                                                  datetime.now())},
                                               'geocoded_data': {}},
                                      'supplied_data': {},
                                      'source': '©OpenStreetMap contributors'
                                      }

    property_not_found: Optional[Dict] = {'meta': {'build': 'NULL', 'status': 200, 'matching': 0,
                                                   'timestamps': {'date': str(date.today()),
                                                                  'time': str(datetime.now()),
                                                                  'official_timestamp': datetime.timestamp(
                                                                      datetime.now())},
                                                   'geocoded_data': {},
                                                   },
                                          'supplied_data': {},
                                          'source': '©OpenStreetMap contributors'}

    edges_error: Optional[Dict] = {'meta': {'build': 'EDGE_ERROR', 'status': 404, 'matching': 0,
                                            'timestamps': {'date': str(date.today()),
                                                           'time': str(datetime.now()),
                                                           'official_timestamp': datetime.timestamp(
                                                               datetime.now())},
                                            'geocoded_data': {}},
                                   'supplied_data': {},
                                   'source': '©OpenStreetMap contributors'
                                   }


def create_dependency_instance(method: Any) -> Any:
    """create dependency through the base utilities class"""
    return PeekGeoUtilities(method)


def check_units(unit_in: str, unit_out: str):
    """check if valid units were inputted into the density model"""

    if unit_in not in density_models.DensitySchema().units:
        raise HTTPException(
            status_code=404,
            detail='Select a valid INPUT unit, one list in the documentation',
            headers={'Error_Description': 'Check Unit in Docs'}
        )
    if unit_out not in density_models.DensitySchema().units:
        raise HTTPException(
            status_code=404,
            detail='Select a valid OUTPUT unit, one list in the documentation',
            headers={'Error_Description': 'Check Unit in Docs'}
        )


################################## API ENDPOINTS BELOW #####################################


def check_radius(radius: int) -> bool:
    """check the radius and return failure response if needed"""

    if radius > GlobalDataModel().max_load_radius:
        return True
    else:
        return False


def radius_response_fail(radius: int, location: list) -> Any:
    """define the failed radius response once exceeded"""
    return {'meta': {'build': 'Radius_EXC', 'status': 505, 'matching': 0,
                     'timestamps': {'date': str(date.today()),
                                    'time': str(datetime.now()),
                                    'official_timestamp': datetime.timestamp(datetime.now())},
                     'geocoded_data': {}},

            'supplied_data': {'latitude': location[0], 'longitude': location[1], 'supplied_radius': radius},
            'source': '©OpenStreetMap contributors'
            }


def success_response(date_results: dict,
                     location: list,
                     radius: int,
                     result_arr: Any):
    """success response creation"""

    return {
        'meta': {'build': 'OK', 'status': 200, 'matching': len(result_arr),
                 'timestamps': {'date': date_results['date'],
                                'time': date_results['time'],
                                'official_timestamp': date_results['official_timestamp']},
                 'geocoded_data': {'load_radius': result_arr,
                                   'certainty': 'Within 10 ft'}},
        'supplied_data': {'latitude': location[0], 'longitude': location[1],
                          'supplied_radius': radius},
        'source': '©OpenStreetMap contributors'
    }


class SetUpLoadRadius:
    """set up the load radius with get_params function"""

    def __init__(self, token: str, coordinates: list,
                 radius: Any, reverse_param: bool = False):
        self.token = token
        self.coordinates = coordinates
        self.radius = radius
        self.reverse_param = reverse_param

    # noinspection PyCompatibility
    def set_up(self):
        """set up computational objects for load radius endpoint"""

        # create backend handlers
        backend_dependency = PeekGeoUtilities(DensityRadius)
        backend_obj = backend_dependency.backend_utility

        # get the global metrics manager object
        metrics_manager = GlobalMetricManager()

        # handle the coordinate input
        location: List = backend_dependency.handle_coordinates(self.coordinates, self.reverse_param)

        # get the call limit
        call_limit = backend_dependency.get_call_limit(self.token, metrics_manager.api_call_key)

        # data packet for dependency
        return [self.radius, location, backend_obj, backend_dependency, self.token, call_limit]


class LoadRadiusContainer(containers.DeclarativeContainer):
    """load radius container"""

    config = providers.Configuration()

    loader = providers.Factory(
        SetUpLoadRadius,
        token=config.token,
        coordinates=config.coordinates,
        radius=config.radius,
        reverse_param=config.reverse_param
    )


# noinspection PyCompatibility
async def load_properties_api(class_dependency: dict) -> dict:
    """functionality/endpoint for load radius tool"""

    # create container
    container = LoadRadiusContainer()
    container.config.from_dict(class_dependency)

    # find all backend objects from dependency
    radius, location, load_radius_backend, utility_obj, token, call_limit = container.loader().set_up()

    # add background task for closing mongo client
    utility_obj.close_client()

    # create call management background task
    await utility_obj.regulate_api_calls(token, 'single')

    # user metric background task here
    await utility_obj.handle_user_metrics('radius', token=token, radius=radius)

    # check api call limit if any calls remain
    if call_limit - 1 >= 0:

        # conduct radius check so users do not exceed
        if check_radius(radius):
            return radius_response_fail(radius, location)

        try:
            # handle coroutines
            task1 = create_task(load_radius_backend.load_radius(location, radius))
            date_task = create_task(utility_obj.get_datetime_objs())

            # process coroutine results
            results = await task1
            date_results = await date_task

            # encoded_json = fast_api_response_packet
            utility_obj.background_api_call_handler(token, 1, call_limit, GlobalMetricManager().api_call_key)

            # create and return the success packet at the bottom
            return success_response(date_results, location, radius, results)

        except BrokenPipeError:
            # 505 server error response
            return NullResponses().internal_error

    else:
        # no calls response
        return NullResponses().exhausted_calls


################################## Convex Search endpoint ########################################


def excessive_edge_length_response(edges: List[list], date_results: dict) -> Dict:
    """edges length exceeded response"""

    return {
        'meta': {'build': 'EDGE_EXC', 'status': 404, 'matching': 0,
                 'timestamps': {'date': date_results['date'],
                                'time': date_results['time'],
                                'official_timestamp': date_results['official_timestamp']},
                 'geocoded_data': {'convex_array': [],
                                   'certainty': 'Within 10 ft'}},
        'supplied_data': {'search_edges': edges},
        'source': '©OpenStreetMap contributors'
    }


# noinspection PyUnresolvedReferences
def count_edge_response(edges: List[list]) -> bool:
    """count the number of edges and check if it exceeds the global maximum set"""
    if len(edges[0]) > GlobalDataModel().convex_edge_limit:
        return True


def duplicate_edge_response(search_edges: List, repeated_edges: List, date_results: dict) -> Dict:
    """duplicate edge response"""
    repeated_edges = [literal_eval(node) for node in set(repeated_edges)]
    return {
        'meta': {'build': 'DUP_EDGE', 'status': 404, 'matching': 0,
                 'timestamps': {'date': date_results['date'],
                                'time': date_results['time'],
                                'official_timestamp': date_results['official_timestamp']},
                 'geocoded_data': {'convex_array': [],
                                   'certainty': 'Within 10 ft',
                                   'repeated_edges': repeated_edges}},
        'geocoded_data': {'search_edges': search_edges[0]},
        'source': '©OpenStreetMap contributors'
    }


def duplicate_edge_check(edges_duplicate: list):
    """check for duplicate edges in user input"""
    duplication_check = check_dups.is_duplicates(edges_duplicate)
    if duplication_check[0]:
        return duplication_check[1]
    else:
        return False


def area_exceeded_response(area: Any, date_results: dict) -> Dict:
    """area exceeded response"""

    return {
        'meta': {'build': 'AREA_EXC, CURRENT: {}'.format(area), 'status': 500, 'matching': 0,
                 'timestamps': {'date': date_results['date'],
                                'time': date_results['time'],
                                'official_timestamp': date_results['official_timestamp']},
                 'geocoded_data': {}},
        'supplied_data': {},
        'source': '©OpenStreetMap contributors'
    }


# noinspection PyUnresolvedReferences
def area_check(class_obj: Callable, latitudes: list, longitudes: list) -> bool:
    """check for area exceeding"""
    area = class_obj.get_polygon_area(latitudes, longitudes)
    if area > GlobalDataModel().convex_area_limit:
        return area
    else:
        return False


def convex_success_response(edges: list, results: Any, date_results: dict):
    """create the success case for the convex endpoint"""

    return {
        'meta': {'build': 'OK', 'status': 200, 'matching': len(list(results)),
                 'timestamps': {'date': date_results['date'],
                                'time': date_results['time'],
                                'official_timestamp': date_results['official_timestamp']},
                 'geocoded_data': {'convex_array': [i for i in results],
                                   'certainty': 'Within 10 ft'}},
        'supplied_data': {'search_edges': edges[0]},
        'source': '©OpenStreetMap contributors'
    }


class SetUpConvex:
    """set up convex class for polygon searches"""

    def __init__(self, token: str, coordinates: list, reverse_param: Optional[bool] = False):
        self.token = token
        self.coordinates = coordinates
        self.reverse_param = reverse_param

    # noinspection PyCompatibility
    def set_up(self):
        """set up all computational objects for the convex endpoint"""

        # create the backend object handlers
        utility_obj = PeekGeoUtilities(Convex)
        backend_obj = utility_obj.backend_utility

        latitudes, longitudes = utility_obj.get_lats_longs(self.coordinates, self.reverse_param)

        # get the call limit
        call_limit = utility_obj.get_call_limit(self.token, GlobalMetricManager().api_call_key)

        # create utility return packet here
        return [backend_obj, utility_obj, self.coordinates, self.token, call_limit, latitudes, longitudes]


class ConvexContainer(containers.DeclarativeContainer):
    """convex container class"""

    config = providers.Configuration()

    loader = providers.Factory(
        SetUpConvex,
        token=config.token,
        coordinates=config.coordinates,
        reverse_param=config.reverse_param
    )


# noinspection PyCompatibility
async def convex_search_api(class_dependency: dict) -> dict:
    """convex search endpoint for all convex functionality"""

    # create the container
    container = ConvexContainer()
    container.config.from_dict(class_dependency)

    # get the api endpoint objects from dependency
    geocoding_library, utility_obj, edges, token, call_limit, latitudes, longitudes = container.loader().set_up()

    # create call management background task
    utility_obj.regulate_api_calls(token, 'single')

    # create date task
    date_task = create_task(utility_obj.get_datetime_objs())

    # await the date_task coroutine
    date_results = await date_task

    # get the inner edges
    inner_edges = edges[0]

    # edges count, duplicate check, and area response here
    if count_edge_response(edges):
        return excessive_edge_length_response(edges, utility_obj)
    duplicate_check = duplicate_edge_check(inner_edges[0:len(inner_edges) - 1])
    if type(duplicate_check) is not bool:
        return duplicate_edge_response(edges, duplicate_check, date_results)
    area_val = area_check(geocoding_library, latitudes, longitudes)
    if type(area_val) is not bool:
        return area_exceeded_response(area_val, date_results)

    # add background tasks for convex endpoint
    utility_obj.handle_user_metrics('convex', token=token,
                                    area=area_val, edge_count=len(edges))

    # create polygon task after verification
    polygon_task = create_task(geocoding_library.convex_director(edges))

    # check if user has api calls available
    if call_limit - 1 >= 0:
        try:
            # await the results from polygon task
            results = await polygon_task
            utility_obj.close_client()

            # decrease api calls only in the success case
            utility_obj.background_api_call_handler(token, 1, call_limit,
                                                    GlobalMetricManager().api_call_key)

            # create and return the success packet here
            return convex_success_response(edges, results, date_results)

        except BrokenPipeError:
            # 505 error response
            utility_obj.close_client()
            return NullResponses().edges_error

    else:
        # no remaining calls response
        return NullResponses().exhausted_calls


#################################### DENSITY SEARCH ENDPOINT HERE ################################################


def exceeded_radius_check(radius: int, unit: str) -> bool:
    """handle exceeded density radius response"""
    if radius > GlobalDataModel().density_radius_limit[unit]:
        return True


def exceeded_radius_response(radius: int, unit_out: str, unit_in: str, coordinates: list,
                             date_results: dict) -> Dict:
    """exceeded radius for density coordinate query"""

    return {
        'meta': {'build': 'RADIUS_EXC'.format(GlobalDataModel().density_radius_limit[unit_in], unit_in), 'status': 404,
                 'matching': 0,
                 'timestamps': {'date': date_results['date'],
                                'time': date_results['time'],
                                'official_timestamp': date_results['official_timestamp']},
                 'geocoded_data': {'density': 0,
                                   'unit': '{}'.format(unit_out)}},
        'supplied_data': {'coordinates': coordinates, 'radius': str(radius)},
        'source': '©OpenStreetMap contributors'
    }


def success_density_response(result: Any, date_results: dict, unit_out: str, supplied: dict):
    """create the success response for the density endpoint"""
    return {'meta': {'build': 'OK', 'status': 200, 'matching': 1,
                         'timestamps': {'date': date_results['date'],
                                        'time': date_results['time'],
                                        'official_timestamp': date_results['official_timestamp']},
                         'geocoded_data': {'density': result,
                                           'unit': '{}'.format(unit_out)}},
                'supplied_data': supplied,
                'source': '©OpenStreetMap contributors'
            }


class SetUpDensity:
    """density base set up class creation"""

    def __init__(self, token: str,
                 coordinate: Optional[Any] = None,
                 unit_in: Optional[str] = "ft",
                 unit_out: Optional[str] = "ft",
                 radius: Optional[Any] = None,
                 reverse_param: Optional[Any] = False,
                 custom_option: Any = None,
                 custom_utility: Any = None
                 ):
        self.token = token
        self.coordinate = coordinate
        self.unit_in = unit_in
        self.unit_out = unit_out
        self.radius = radius
        self.reverse_param = reverse_param
        self.custom_option = custom_option
        self.custom_utility = custom_utility

    # noinspection PyCompatibility
    def set_up(self):
        """set up all computational objects for the density endpoint"""

        # create the backend handlers
        utility_obj = PeekGeoUtilities(DensityRadius)
        backend_obj = utility_obj.backend_utility

        # check the strategies and translate into usable models
        if self.custom_option is not None and self.custom_utility is not None:
            strategy = utility_obj.custom_density_mapping[utility_obj.runner_model.engine][self.custom_option]
        else:
            strategy = None

        # handle the user coordinate
        coordinate = utility_obj.handle_coordinates(self.coordinate, self.reverse_param)

        # make sure input unit is supported and raise HTTPException if not
        check_units(self.unit_in, self.unit_out)

        # check if there is any input at all
        if coordinate == [None, None] and strategy is None:
            raise HTTPException(
                status_code=404,
                detail='No Input Detected',
                headers={'Error_Description': "Fill Required Input Fields"}
            )

        # get call limit
        call_limit = utility_obj.get_call_limit(self.token, GlobalMetricManager().api_call_key)
        return [backend_obj, utility_obj, coordinate, self.radius, self.unit_in, self.unit_out,
                self.custom_utility, strategy, self.token, call_limit]


class DensityContainer(containers.DeclarativeContainer):
    """density container class for density functionality support"""

    # create config container for variable storage
    config = providers.Configuration()

    # create dependency instance
    loader = providers.Factory(
        SetUpDensity,
        token=config.token,
        coordinate=config.coordinate,
        unit_in=config.unit_in,
        unit_out=config.unit_out,
        radius=config.radius,
        reverse_param=config.reverse_param,
        custom_option=config.custom_option,
        custom_utility=config.custom_utility
    )


# noinspection PyCompatibility
async def density_search_api(class_dependency: dict) -> dict:
    """density search endpoint functionality"""

    # create container instance
    container = DensityContainer()
    container.config.from_dict(class_dependency)

    fn, utility_obj, coordinates, radius, unit_in, unit_out, custom_query_value, strategy, token, call_limit = container.loader().set_up()

    # create call management background task
    await utility_obj.regulate_api_calls(token, 'single')

    date_task = create_task(utility_obj.get_datetime_objs())
    date_results = await date_task

    # check the call exceeding condition
    if call_limit - 1 >= 0:

        # deal with custom density search first
        if strategy is not None:
            result = create_task(fn.density_factory("custom", strategy=strategy, query_value=custom_query_value))
            result = await result
            utility_obj.close_client()

            # add user metrics
            await utility_obj.handle_user_metrics('density', token=token, radius=radius,
                                            unit_in=unit_in, unit_out=unit_out, density=result, custom=True)

            return success_density_response(result, date_results, unit_out, {'custom_utility': custom_query_value})

        # check if inputted radius exceeds the global variable set at top
        if exceeded_radius_check(radius, unit_in):
            return exceeded_radius_response(radius, unit_out, unit_in, coordinates, date_results)

        try:
            # create coroutines here
            result = await create_task(
                fn.density_factory("generic", coordinates=coordinates, radius=radius, unit_in=unit_in))
            utility_obj.close_client()

            await utility_obj.handle_user_metrics('density', token=token, radius=radius,
                                            unit_in=unit_in, unit_out=unit_out, density=result, custom=False)

            # handle API calls
            utility_obj.background_api_call_handler(token, 1, call_limit,
                                                    GlobalMetricManager().api_call_key)

            return success_density_response(result, date_results, unit_out, {"radius": radius, "coordinates": coordinates})

        # any error occurs
        except BrokenPipeError:
            # add user metrics
            utility_obj.handle_user_metrics('density', token=token, radius=radius,
                                            unit_in=unit_in, unit_out=unit_out, density='ISE')
            utility_obj.close_client()
            return NullResponses().internal_error

    else:
        # null calls response
        return NullResponses().exhausted_calls


########################### REVERSE GEOCODER ###################################


def success_single_reverse(date_results: dict, formatted_address: str, location: list,
                           radius: Any, address_components: dict, confidence: float):
    """create success case singular reverse geocoding response"""

    return {'meta': {'build': 'OK', 'status': 200, 'matching': 1,
                     'timestamps': {'date': date_results['date'],
                                    'time': date_results['time'],
                                    'official_timestamp': date_results['official_timestamp']},
                     'geocoded_data': {'format_address': formatted_address,
                                       'address_parts': address_components,
                                       'address_validation': True,
                                       'confidence': round(confidence, 1)}},
            'supplied_data': {'coordinates': location, 'radius': radius},
            'source': '©OpenStreetMap contributors'
            }


def poi_not_found_response(date_results: dict, formatted_address: Any, radius: int,
                           address_components: dict, location: list) -> Dict:
    """not found response for place of interest"""

    return {'meta': {'build': 'POI_NULL', 'status': 200, 'matching': 0,
                     'timestamps': {'date': date_results['date'],
                                    'time': date_results['time'],
                                    'official_timestamp': date_results['official_timestamp']},
                     'geocoded_data': {'format_address': formatted_address,
                                       'address_parts': address_components,
                                       'address_validation': False,
                                       'confidence': 0.0}},
            'supplied_data': {'coordinates': location, 'radius': radius},
            'source': '©OpenStreetMap contributors'
            }


class SetUpReverse:
    """radius setup class for basic functionalities"""

    def __init__(self, token: str,
                 coordinate: list,
                 radius: Any = 10,
                 reverse_param: Optional[bool] = False
                 ):
        self.token = token
        self.coordinate = coordinate
        self.radius = radius
        self.reverse_param = reverse_param

    # noinspection PyCompatibility
    def set_up(self):
        """set up function for singular reverse geocoding endpoint"""

        # create backend object handlers
        utility_obj = PeekGeoUtilities(GeneralReverseManager)
        backend_obj = utility_obj.backend_utility

        coordinate = utility_obj.handle_coordinates(self.coordinate, self.reverse_param)

        # get the call limit
        call_limit = utility_obj.get_call_limit(self.token, GlobalMetricManager().api_call_key)

        # define response array
        return [backend_obj, utility_obj, coordinate, self.radius, call_limit, self.token]


class ReverseContainer(containers.DeclarativeContainer):
    """reverse container class creation"""

    # create configuration variable storage
    config = providers.Configuration()

    # create running instance of radius set up class
    loader = providers.Factory(
        SetUpReverse,
        token=config.token,
        coordinate=config.coordinate,
        radius=config.radius,
        reverse_param=config.reverse_param
    )


# noinspection PyCompatibility
async def reverse_geocode_api(class_dependency: dict) -> dict:
    """endpoint for single reverse geocoding requests"""

    # create container instance
    container = ReverseContainer()
    container.config.from_dict(class_dependency)

    # get set-up objects
    geocoding_library, utility_obj, location, radius, call_limit, token = container.loader().set_up()

    # add background task for closing mongo client
    utility_obj.close_client()

    # create call management background task
    await utility_obj.regulate_api_calls(token, 'single')

    # create datetime objects coroutine
    date_results = create_task(utility_obj.get_datetime_objs())

    # add user metrics
    await utility_obj.handle_user_metrics('reverse_single',
                                          coordinates=location, token=token)

    # check if API call limit was exceeded
    if call_limit - 1 >= 0:

        try:
            # client code request for backend reverse-geocoded data
            formatted_address, address_components, confidence = geocoding_library.single_reverse(location, radius)

            # get the date results from coroutine
            date_results = await date_results

            # check if the property response is empty (when poi is not found)
            if common_models.poi_not_found_check(formatted_address):
                return poi_not_found_response(date_results, formatted_address, radius, address_components, location)

            # deduct one api call from user profile
            utility_obj.background_api_call_handler(token, 1, call_limit,
                                                    GlobalMetricManager().api_call_key)

            return success_single_reverse(date_results, formatted_address, location, radius, address_components,
                                          confidence)

        except BrokenPipeError:
            # 505 internal error
            return NullResponses().property_not_found

    else:
        # null call response
        return NullResponses().exhausted_calls


############################### BATCH REVERSE GEO ENDPOINT ##################################


def batch_reverse_success(date_results: Dict, geocoded_data: List[Any], coordinates: List[list], radius: int):
    """create success response for batch reverse geocoding endpoint"""
    return {'meta': {'build': 'OK', 'status': 200, 'matching': len(geocoded_data),
                     'timestamps': {'date': date_results['date'],
                                    'time': date_results['time'],
                                    'official_timestamp': date_results['official_timestamp']},
                     'geocoded_data': geocoded_data},
            'supplied_data': {'coordinates': coordinates, 'radius': radius},
            'source': '©OpenStreetMap contributors'
            }


def batch_reverse_duplicate(date_results: Dict, repeated_edges, coordinates: List[list], radius: int):
    """create success response for batch reverse geocoding endpoint"""
    repeated_edges = [literal_eval(node) for node in repeated_edges]
    return {'meta': {'build': 'DUP', 'status': 400, 'matching': 0,
                     'timestamps': {'date': date_results['date'],
                                    'time': date_results['time'],
                                    'official_timestamp': date_results['official_timestamp']},
                     'geocoded_data': [{"repeated_edges": repeated_edges}]},
            'supplied_data': {'coordinates': coordinates, 'radius': radius},
            'source': '©OpenStreetMap contributors'
            }


class SetUpBatchReverse:
    """setup class for batch reverse geocoding"""

    def __init__(self, token: str,
                 latitudes: Optional[List[float]] = Query(None),
                 longitudes: Optional[List[float]] = Query(None),
                 radius: Any = 10,
                 coordinate_file: Optional[UploadFile] = File(None),
                 reverse_param: Optional[bool] = False
                 ):
        self.token = token
        self.latitudes = latitudes
        self.longitudes = longitudes
        self.radius = radius
        self.coordinate_file = coordinate_file
        self.reverse_param = reverse_param

    # noinspection PyCompatibility
    def set_up(self):
        """batch reverse geocoding functionality"""

        # create backend handlers
        utility_obj = PeekGeoUtilities(GeneralReverseManager)
        backend_obj = utility_obj.backend_utility

        # create the coordinates array regardless of File(...) input type
        coordinates = utility_obj.handle_batch_reverse_coords(self.latitudes, self.longitudes, self.reverse_param,
                                                              self.coordinate_file)

        # get call limit for user
        call_limit = utility_obj.get_call_limit(self.token, GlobalMetricManager().api_call_key)

        return [self.coordinate_file, backend_obj, utility_obj, "None", coordinates, self.reverse_param, self.radius,
                self.token, call_limit]


class BatchReverseContainer(containers.DeclarativeContainer):
    """create container instance for batch reverse geocoding"""

    # create configuration
    config = providers.Configuration()

    # running instance for setup class
    loader = providers.Factory(
        SetUpBatchReverse,
        token=config.token,
        latitudes=config.latitudes,
        longitudes=config.longitudes,
        radius=config.radius,
        coordinate_file=config.coordinate_file,
        reverse_param=config.reverse_param
    )


# noinspection PyCompatibility
async def batch_reverse_api(class_dependency: dict) -> dict:
    """batch reverse geocoding endpoint"""

    # create container
    container = BatchReverseContainer()
    container.config.from_dict(class_dependency)

    # get objects used for computation
    coordinate_file, geocoding_library, utility_obj, user_agent, coordinates, reverse_param, radius, token, call_limit = container.loader().set_up()

    # get the filename for logging purposes
    file_name = os.path.basename(__file__)

    # create call management background task
    await utility_obj.regulate_api_calls(token, 'single')

    # make date results coroutine
    dates = create_task(utility_obj.get_datetime_objs())

    # handle date coroutine
    date_results = await dates

    # check duplicate coordinates
    duplicate_check = duplicate_edge_check(coordinates)
    if type(duplicate_check) is not bool:
        return batch_reverse_duplicate(date_results, duplicate_check, coordinates, radius)

    # user metrics
    await utility_obj.handle_user_metrics('reverse_batch', token=token,
                                    coordinates=coordinates, batch_size=len(coordinates))

    if call_limit - 1 >= 0:
        try:
            # data from the batch geocoder in backend is returned
            data = geocoding_library.batch_reverse(coordinates, radius)
            utility_obj.close_client()

            # handle api calls, subtract
            utility_obj.background_api_call_handler(token, 1, call_limit,
                                                    GlobalMetricManager().api_call_key)

            # logging for user data
            await utility_obj.background_logging('AddUserLogger', '/Users/martinmashalov/Documents/Python/PeekGeo/Logs/BatchReverse.txt',
                                           'DEBUG','USER-AGENT: {}, Coordinates: {}, Arr_Len: {}, results: {}, file: {}'.format(
                                               user_agent, coordinates, len(coordinates), data[0], file_name))

            return batch_reverse_success(date_results, data, coordinates, radius)

        # internal server error
        except BrokenPipeError or BufferError:
            utility_obj.background_logging('AddUserLogger',
                                           '/Users/martinmashalov/Documents/Python/PeekGeo/Logs/ReverseGeocodingBatch.txt',
                                           'CRITICAL', 'USER-AGENT: {}, Buffer Error encountered'.format(user_agent))
            utility_obj.close_client()
            return NullResponses().internal_error

    else:
        # null call response
        return NullResponses().exhausted_calls


#################################### Single Geocoding Endpoint ########################################


def create_geocode_singe_success(date_results: dict, response_packet: list, supplied_packet: list) -> Dict:
    """create success response packet for singular geocoder"""
    return {'meta': {'build': 'OK', 'status': 200, 'matching': 1,
                     'timestamps': {'date': date_results['date'],
                                    'time': date_results['time'],
                                    'official_timestamp': date_results['official_timestamp']
                                    },
                     'geocoded_data': response_packet
                     },
            'supplied_data': supplied_packet,
            'source': '©OpenStreetMap contributors'
            }


class SetUpSingleForward:
    """create setup instance for forward geocoding function"""

    def __init__(self, token: str,
                 full_address: Optional[str] = None,
                 street: Optional[str] = None,
                 number: Optional[int] = None,
                 postcode: Optional[int] = None,
                 region: Optional[str] = None,
                 city: Optional[str] = None,
                 unit: Optional[str] = None,
                 ):
        self.token = token
        self.full_address = full_address
        self.street = street
        self.number = number
        self.postcode = postcode
        self.region = region
        self.city = city
        self.unit = unit

    # noinspection PyCompatibility
    def set_up(self):
        """singular geocoding computational object creation"""

        # backend object handlers
        utility_obj = PeekGeoUtilities(GeneralGeocodingManager)
        backend_obj = utility_obj.backend_utility

        # get the call limit
        call_limit = utility_obj.get_call_limit(self.token, GlobalMetricManager().api_call_key)

        # if address is None, then use the parsed input for address
        if self.full_address is None:
            # create parsed dictionary, with checks for accuracy
            try:
                if self.region.upper() in common_models.CommonUtilityModel().states and len(str(self.postcode)) == 5:
                    self.full_address: Dict = dict(street=self.street, number=self.number, state=self.region,
                                              postcode=self.postcode, city=self.city,
                                              unit=self.unit)
                    address_type = 0
                else:
                    raise HTTPException(
                        status_code=500,
                        detail='State must follow the abbreviations in the documentation; postcode must be a valid 5 digit postcode',
                        headers={"Error_Description": "Check type and formatting of postcode and street inputs"}
                    )
            except AttributeError:
                raise HTTPException(
                    status_code=404,
                    detail='Must include either parsed address or full address. No Input Detected',
                    headers={"Error_Description": "Fulfill address formatting requirements, check docs."}
                )

        # just return the full address string
        else:
            address_type = 1

        return [backend_obj, utility_obj, self.full_address, address_type, call_limit, self.token]


class SingleGeocodeContainer(containers.DeclarativeContainer):
    """create single geocoder container"""

    # create configuration
    config = providers.Configuration()

    # running instance of setup base class
    loader = providers.Factory(
        SetUpSingleForward,
        token=config.token,
        full_address=config.full_address,
        street=config.street,
        number=config.number,
        postcode=config.postcode,
        region=config.region,
        city=config.city,
        unit=config.unit
    )


# noinspection PyCompatibility
async def single_geocode_api(class_dependency: dict) -> dict:
    """singular geocoding endpoint"""

    # create container
    container = SingleGeocodeContainer()
    container.config.from_dict(class_dependency)

    # get computational objects
    geocoding_library, utility_obj, address, address_type, call_limit, token = container.loader().set_up()

    # create call management background task
    await utility_obj.regulate_api_calls(token, 'single')

    # create datetime coroutine
    dates = create_task(utility_obj.get_datetime_objs())

    # check the call limitations
    if call_limit - 1 >= 0:
        try:
            # add user metrics
            await utility_obj.handle_user_metrics('forward_single', token=token,
                                            address=address, full=0, parsed=1)

            # get the output
            output_packets = geocoding_library.singular_geocode(address, address_type)
            utility_obj.close_client()

            # await date coroutine
            date_results = await dates

            # subtract an api call in the background
            utility_obj.background_api_call_handler(token, 1, call_limit,
                                                    GlobalMetricManager().api_call_key)

            # make the response packet
            return create_geocode_singe_success(date_results, output_packets['response_packet'],
                                                output_packets['supplied_packet'])

        except ValueError:
            # 505 internal server response
            utility_obj.close_client()
            return NullResponses().internal_error

    else:
        # null call response
        return NullResponses().exhausted_calls


################################## Batch Geocoder Endpoint ####################################


def success_batch_forward(date_results, geocoded_data: dict) -> Dict:
    """create success case response packet for batch geocoding"""

    return {'meta': {'build': 'OK', 'status': 200, 'matching': len(geocoded_data['response_packet']),
                     'timestamps': {'date': date_results['date'],
                                    'time': date_results['time'],
                                    'official_timestamp': date_results['official_timestamp']
                                    },
                     'geocoded_data': geocoded_data['response_packet']
                     },
            'supplied_data': geocoded_data['supplied_packet'],
            'source': '©OpenStreetMap contributors'
            }


class SetUpBatchGeocode:
    """batch geocode setup base class"""

    def __init__(self, token: str,
                 addresses: List[str] = Query(None),
                 address_file: Optional[UploadFile] = File(None)
                 ):
        self.token = token
        self.addresses = addresses
        self.address_file = address_file

    # noinspection PyCompatibility
    def set_up(self):
        """batch geocoding backend code for FastAPI"""

        # backend object handlers
        utility_obj = PeekGeoUtilities(GeneralGeocodingManager)
        backend_obj = utility_obj.backend_utility

        # get call limit
        call_limit: int = utility_obj.get_call_limit(self.token, GlobalMetricManager().api_call_key)

        # address processing
        address_arr: list = utility_obj.handle_batch_reverse_addresses(self.addresses, self.address_file)

        return [backend_obj, utility_obj, address_arr,
                self.address_file.filename if self.address_file is not None else None,
                call_limit, self.token]


class BatchGeocodeContainer(containers.DeclarativeContainer):
    """container for batch reverse geocoding"""

    # create configuration for variable storage
    config = providers.Configuration()

    # running instance of base setup class
    loader = providers.Factory(
        SetUpBatchGeocode,
        token=config.token,
        addresses=config.addresses,
        address_file=config.address_file
    )


# noinspection PyCompatibility
async def batch_geocode_api(class_dependency: dict) -> dict:
    """batch geocoding endpoint function"""

    # create container
    container = BatchGeocodeContainer()
    container.config.from_dict(class_dependency)

    # create class object and other computational objects
    backend_obj, utility_obj, addresses, file_name, call_limit, token = container.loader().set_up()

    # create call management background task
    await utility_obj.regulate_api_calls(token, 'single')

    # create date_results coroutine
    date_results = create_task(utility_obj.get_datetime_objs())

    # add user metrics
    await utility_obj.handle_user_metrics('forward_batch', token=token,
                                    addresses=addresses,
                                    batch_size=len(addresses))

    # check call limitations
    if call_limit - 1 >= 0:
        # get date results here
        date_results = await date_results

        try:
            # get backend data
            data = backend_obj.batch_geocode(addresses)
            utility_obj.close_client()

            # add background task for call limit handling
            utility_obj.background_api_call_handler(token, 1, call_limit,
                                                    GlobalMetricManager().api_call_key)

            return success_batch_forward(date_results, data)

        except BrokenPipeError or BufferError:
            # internal error response
            utility_obj.close_client()
            return NullResponses().internal_error

    else:
        # null call response
        return NullResponses().exhausted_calls


############################### MAIN FACADE CLASS ###############################


class places:
    """places facade class for package contents"""

    def __init__(self, token: str):
        self.token: str = token

    @staticmethod
    def process_basic_args(additional_args: dict) -> bool:
        """process the basic arguments for load radius, convex, density, reverse, and batch reverse methods"""

        # reverse_parameter extraction
        try:
            reverse_param: Any = additional_args['reverse_param']
        except KeyError:
            reverse_param: Any = None

        return reverse_param

    def load_properties(self, coordinates: list = (), radius: Any = 10, **additionals):
        """load properties facade for package purposes"""

        # get token and reverse parameter
        reverse_param = self.process_basic_args(additionals)

        # create and send response package
        return run(load_properties_api(
            {'coordinates': coordinates, 'radius': radius, 'token': self.token, 'reverse_param': reverse_param}))

    def convex(self, coordinates: list, **additionals) -> Any:
        """convex search facade method for places package"""

        # get token and reverse parameter
        reverse_param = self.process_basic_args(additionals)

        # create response and send it
        return convex_search_api({"token": self.token, "coordinates": coordinates, "reverse_param": reverse_param})

    def density(self, unit_in="ft", unit_out="ft", coordinate: list = None, radius: Any = 1,
                custom_option: str = None, custom_utility: Any = None, **additionals) -> Any:
        """density facade function for places package"""

        # get token and reverse parameter
        reverse_param = self.process_basic_args(additionals)

        return run(density_search_api(
            {"token": self.token, "coordinate": coordinate, "unit_in": unit_in, "unit_out": unit_out,
             "radius": radius, "reverse_param": reverse_param, "custom_option": custom_option,
             "custom_utility": custom_utility}))

    def reverse(self, coordinate: list = (), radius: Any = 10, **additionals) -> Any:
        """reverse geocoding facade method for Places package"""
        # get token and reverse parameter
        reverse_param = self.process_basic_args(additionals)

        # create and send response packet
        return run(reverse_geocode_api({"token": self.token, "coordinate": coordinate, "radius": radius,
                                        "reverse_param": reverse_param}))

    def batch_reverse(self, coordinates: List[list] = (()), radius: Any = 10, coordinate_file: UploadFile = None,
                      **additionals) -> Any:
        """batch reverse geocoding facade method for Places package"""

        # get token and reverse parameter
        reverse_param = self.process_basic_args(additionals)

        # create containers for lats and longs
        lats, longs = coordinate_conv_package.coords_conv(coordinates)

        # create and send response packet
        return run(batch_reverse_api({"token": self.token, "latitudes": lats, "longitudes": longs, "radius": radius,
                                  "coordinate_file": coordinate_file, "reverse_param": reverse_param}))

    def forward(self, full_address: str = None, street: str = None, number: Any = None, postcode: int = None, region: str = None, city:
                      str = None, unit=None) -> Any:
        """facade method for forward geocoding on Places API"""

        return run(single_geocode_api(
            {"token": self.token, "full_address": full_address, "street": street, "number": number,
             "postcode": postcode, "region": region, "city": city, "unit": unit}))

    def batch_forward(self, addresses: list = (), address_file: UploadFile = None) -> Any:
        """batch forward facade method"""

        # create the send response
        return run(batch_geocode_api({"token": self.token, "addresses": addresses, "address_file": address_file}))


__all__ = [
    places
]



import requests
import pymongo
import ssl
import datetime
import time
import ast


# TODO: complete testing code for API endpoints
# TODO: debug any issues and make sure it works spotless
###DEFINE timing function at the top to used as a decorator
def time_function(function):
    def wrapper_function(*args, **kwargs):
        start = time.perf_counter()
        function(*args, **kwargs)
        total_runtime = time.perf_counter() - start
        return total_runtime

    return wrapper_function


###CLASS definitions below for unit testing capability
class SetUpDBS:

    def __init__(self, dataset, maxSize):
        self.dataset = dataset
        self.maxSize = maxSize
        self.mongo_pass = 'Barca2004!'

        self.client = pymongo.MongoClient(
            'mongodb+srv://martin_mashalov:{}@locationpoints.wi1dk.mongodb.net/test'.format(self.mongo_pass),
            maxPoolSize=maxSize,
            ssl_cert_reqs=ssl.CERT_NONE)

        self.db = self.client[self.dataset]
        self.property_collection = self.db['propLocations']
        self.account_collection = self.db['Accounts']

    def __len__(self):
        return self.account_collection.estimated_document_count()

    def __float__(self, input_case):
        if isinstance(input_case, int) or isinstance(input_case, float):
            return float(input_case)
        else:
            raise Exception("Uncorrectable Type Error")

    def __call__(self, first, last):
        account = self.account_collection.find({'$and': [{'details.firstname': first}, {'details.lastname': last}]})
        try:
            account = [i for i in account][0]
            return len(account['preferences']['properties'])
        except:
            return 0

class PeekAppEndpointTest(SetUpDBS):

    def __init__(self, dataset):
        self.maxpoolsize = 50
        self.dataset = dataset
        self.coordinates = [43.0789363, -77.6462623]
        self.property_ids = []
        super(PeekAppEndpointTest, self).__init__(self.dataset, self.maxpoolsize)

        self.property_urls = {
            'GetPropertyDetails': 'http://127.0.0.1:5000/get_property/?availability=all&house_loc=[43.0789363, -77.6462623]',
            'LoadCoordinates': 'http://127.0.0.1:5000/load_properties/?location=[43.0785005, -77.6459917]&radius=1000'
        }

        self.user_urls = {
            'AddUserProperty1': 'http://realstate-dev.us-east-1.elasticbeanstalk.com/add_user_property/?lastname=asdfjajfhksjhfkajs&firstname=uqowasfhsdfh&email=abc1@gmail.com&addr=4994 State Route 23, Windham, NY, 12494&price=3450001&bed=3&bath=3&tax=2035&area=1812&pool=no&stat=available&image=https://ap.rdcpix.com/e6588db17b627c293a40dc443dad00afl-m3631576332x.jpg, https://ar.rdcpix.com/e6588db17b627c293a40dc443dad00afl-m3631576332x.jpg&lat=42.306199&long=-74.223715&date_up=05/01/2021&year=2005',
            'AddUserProperty2': 'http://realstate-dev.us-east-1.elasticbeanstalk.com/add_user_property/?lastname=asdfjajfhksjhfkajs&firstname=uqowasfhsdfh&email=abc1@gmail.com&addr=4995 State Route 23, Windham, NY, 12494&price=3450001&bed=3&bath=3&tax=2035&area=1812&pool=no&stat=available&image=https://ap.rdcpix.com/e6588db17b627c293a40dc443dad00afl-m3631576332x.jpg, https://ar.rdcpix.com/e6588db17b627c293a40dc443dad00afl-m3631576332x.jpg&lat=42.306199&long=-74.223715&date_up=05/01/2021&year=2005',
            'AddUserSameEmailError': 'http://realstate-dev.us-east-1.elasticbeanstalk.com/add_user_property/?lastname=asdfjajfhksjhfkajs&firstname=owasfhsdfh&email=abc1@gmail.com&addr=4996 State Route 23, Windham, NY, 12494&price=3450001&bed=3&bath=3&tax=2035&area=1812&pool=no&stat=available&image=https://ap.rdcpix.com/e6588db17b627c293a40dc443dad00afl-m3631576332x.jpg, https://ar.rdcpix.com/e6588db17b627c293a40dc443dad00afl-m3631576332x.jpg&lat=42.306199&long=-74.223715&date_up=05/01/2021&year=2005',
            'AddUserDifEmailSameName': 'http://realstate-dev.us-east-1.elasticbeanstalk.com/add_user_property/?lastname=asdfjajfhksjhfkajs&firstname=uqowasfhsdfh&email=abc123@gmail.com&addr=4995 State Route 23, Windham, NY, 12494&price=3450001&bed=3&bath=3&tax=2035&area=1812&pool=no&stat=available&image=https://ap.rdcpix.com/e6588db17b627c293a40dc443dad00afl-m3631576332x.jpg, https://ar.rdcpix.com/e6588db17b627c293a40dc443dad00afl-m3631576332x.jpg&lat=42.306199&long=-74.223715&date_up=05/01/2021&year=2005',
            'GetFavoriteList': 'http://realstate-dev.us-east-1.elasticbeanstalk.com/retrieve_user_data/?email=abc1@gmail.com',
            'DelProperty': 'http://realstate-dev.us-east-1.elasticbeanstalk.com/del_property/?firstname=uqowasfhsdfh&lastname=asdfjajfhksjhfkajs&email=abc1@gmail.com&property_id='}

        self.query_collection = {
            'primary_user': {'$and': [{'details.firstname': 'uqowasfhsdfh'},
                                      {'details.lastname': 'asdfjajfhksjhfkajs'},
                                      {'details.email': 'abc1@gmail.com'}]},
            'secondary_user': {'$and': [{'details.firstname': 'uqowasfhsdfh'},
                                        {'details.lastname': 'asdfjajfhksjhfkajs'},
                                        {'details.email': 'abc123@gmail.com'}]}
        }

    # should say failed
    @staticmethod
    def AddUser_sameEmail(url):

        MetaClass = SetUpDBS('PeekDB', 50)
        account_db_length = len(MetaClass)

        with open('UnitTestFile.txt', 'a+') as unit_test_file:
            try:
                status = requests.head(url).status_code
                if 200 < status <= 300:
                    status_code = 'GOOD'
                    data = requests.get(url).json()
                    message = data['msg']
                    code = data['code']
                else:
                    status_code = 'ERROR/WRONG ID'
                    message = 'no duplicate detected, should have'
                    code = 0

            except ValueError:
                status_code = 'ERROR/WRONG ID'
                message = 'no duplicated detected, should have'
                code = 0

            new_account_db_length = len(MetaClass)
            if new_account_db_length == account_db_length:
                duplicate_added = 'NO DUP ADDED'
            else:
                duplicate_added = 'DUP ADDED, CRITICAL, SAME EMAIL'

            return 'Testing: {}; URL status(raw): {} URL status(interpretted): {} msg_line_endpoint: {} function_code: {} UserAdded?: {}'.format(
                'AddUser(SameEmailErrorCheck)',
                status,
                status_code,
                message,
                code,
                duplicate_added)

    # should say failed, use url AddUserEndpoint<1>
    @staticmethod
    def AddProp_dup(url):
        MetaClass = SetUpDBS('PeekDB', 50)
        init_favorite_len = MetaClass('uqowasfhsdfh', 'asdfjajfhksjhfkajs')

        with open('UnitTestFile.txt', 'a+') as unit_test_file:
            try:
                status = requests.head(url).status_code
                if 200 < status <= 300:
                    status_code = 'GOOD'
                    data = requests.get(url).json()
                    message = data['msg']
                    code = data['code']
                else:
                    status_code = 'ERROR/WRONG ID'
                    message = 'no duplicate detected, should have'
                    code = 0

            except ValueError:
                status_code = 'ERROR/WRONG ID'
                message = 'no duplicated detected, should have'
                code = 0

            new_favorite_len = MetaClass('uqowasfhsdfh', 'asdfjajfhksjhfkajs')
            if init_favorite_len == new_favorite_len:
                duplicate_added = 'NO DUP ADDED'
            else:
                duplicate_added = 'DUP ADDED, CRITICAL'

            return 'Testing: {}; URL status(raw): {} URL status(interpretted): {} msg_line_endpoint: {} function_code: {} UserAdded?: {}'.format(
                'AddUserEndpoint(Duplicate property addition test)',
                status,
                status_code,
                message,
                code,
                duplicate_added)

    # should add the property and succeed
    @staticmethod
    def AddUser_SameNameDifEmail(url):
        MetaClass = SetUpDBS('PeekDB', 50)
        init_account_len = len(MetaClass)

        try:
            status = requests.head(url).status_code
            print(status, 'status of AddUser_SameNameDifEmail request')
            if 200 < status <= 300:
                status_code = 'GOOD'
                data = requests.get(url).json()
                message = data['msg']
                code = data['code']
            else:
                status_code = 'ERROR/WRONG ID'
                message = 'no duplicate detected, should have'
                code = 0

        except ValueError:
            status = 400
            status_code = 'ERROR/WRONG ID'
            message = 'no duplicated detected, should have'
            code = 0

        new_account_len = len(MetaClass)
        if init_account_len == new_account_len - 1:
            duplicate_added = 'NEW USER ADDED'
        else:
            duplicate_added = 'NO USER ADDED, CRITICAL'

        return 'Testing: {}; URL status(raw): {} URL status(interpretted): {} msg_line_endpoint: {} function_code: {} UserAdded?: {}'.format(
            'AddUserEndpoint(Add another user endpoint)',
            status,
            status_code,
            message,
            code,
            duplicate_added)

    def remove_added_test_users(self, **kwargs):

        try:
            if kwargs['user_opt'] == 1:
                query = self.account_collection.find(self.query_collection['primary_user'])
                query = [i for i in query][0]

                user_id = query['details']['user_id']
                property_ids = [i['property_id'] for i in query['preferences']['properties']]

                self.account_collection.delete_one({'$and': [{'details.firstname': 'uqowasfhsdfh'},
                                                             {'details.lastname': 'asdfjajfhksjhfkajs'},
                                                             {'details.email': 'abc1@gmail.com'}]})

                self.account_collection.update_one({'Name.n': 'IDS_CHECK'}, {'$pull': {'ID_collections': user_id}})
                for prop_id in property_ids:
                    self.account_collection.update_one({'Name.n': 'IDS_CHECK'}, {'$pull': {'Property_ID_Arr': prop_id}})

            elif kwargs['user_opt'] == 2:
                query = self.account_collection.find(self.query_collection['secondary_user'])
                query = [i for i in query][0]

                user_id = query['details']['user_id']
                property_ids = [i['property_id'] for i in query['preferences']['properties']]

                self.account_collection.delete_one({'$and': [{'details.firstname': 'uqowasfhsdfh'},
                                                             {'details.lastname': 'asdfjajfhksjhfkajs'},
                                                             {'details.email': 'abc123@gmail.com'}]})

                self.account_collection.update_one({'Name.n': 'IDS_CHECK'}, {'$pull': {'ID_collections': user_id}})
                for prop_id in property_ids:
                    self.account_collection.update_one({'Name.n': 'IDS_CHECK'}, {'$pull': {'Property_ID_Arr': prop_id}})


        except ValueError or TypeError or SystemError:
            raise Exception('The removal of test users accounted an unexpected system issue.')

    def get_property_ids(self, *type_arg, **user_kwarg):

        try:
            if isinstance(type_arg[0], list):
                if user_kwarg['user_id'].lower() == 'primary':
                    query = self.account_collection.find(self.query_collection['primary_user'])
                else:
                    query = self.account_collection.find(self.query_collection['secondary_user'])
                query = [i for i in query][0]

                for property in query['preferences']['properties']:
                    if isinstance(self.property_ids, list):
                        self.property_ids.append(property['property_id'])
                    else:
                        continue

            elif isinstance(type_arg[0], tuple):
                if user_kwarg['user_id'].lower() == 'primary':
                    query = self.account_collection.find(self.query_collection['primary_user'])
                else:
                    query = self.account_collection.find(self.query_collection['secondary_user'])
                query = [i for i in query][0]

                for property in query['preferences']['properties']:
                    if isinstance(self.property_ids, list):
                        self.property_ids.append(property['property_id'])
                    else:
                        continue

                self.property_ids = tuple(self.property_ids)

            else:
                raise TypeError

        except:
            raise Exception('Exception arise within get_prop_ids.')

    def check_realtor_key_presence(self, **kwargs):

        # query the database for the coordinate and BSON -> JSON conversion completed here
        def query_database():
            query = self.property_collection.find({'location.coordinates': self.coordinates})
            query = [i for i in query][0]
            return query

        if kwargs['time'].lower() == 'before' or kwargs['time'].lower() == 'b':
            query_result = query_database()
            if query_result['properties']['Realtor_ID'] == 0 or isinstance(query_result['properties']['Realtor_ID'],
                                                                           int):
                API_status = 1
                return API_status
            else:
                API_status = 0
                return API_status

        elif kwargs['time'].lower() == 'after' or kwargs['time'].lower() == 'a':
            query_result = query_database()
            if query_result['properties']['Realtor_ID'] != 0 or isinstance(query_result['properties']['Realtor_ID'],
                                                                           int):
                API_status = 1
                return API_status
            else:
                API_status = 0
                return API_status

        else:
            raise Exception('Improper **kwargs option set')

    def test_user_endpoints(self):
        with open('UnitTestFile.txt', 'a+') as unit_test_file:

            unit_test_file.write('--------- Peek Full Unit Test: {} ---------'.format(datetime.datetime.now()))
            unit_test_file.write('\n')

            MetaClass = SetUpDBS('PeekDB', 50)
            account_db_length = len(MetaClass)

            api_DBreplacement_status = self.check_realtor_key_presence(time='b')

            # try add_user_endpoint here
            try:
                status = requests.head(self.user_urls['AddUserProperty1']).status_code
                if 200 <= status < 300:
                    status_code = 'GOOD'
                    data = requests.get(self.user_urls['AddUserProperty1']).json()
                    message = data['msg']
                    property_id = data['data']['property_id']
                else:
                    status_code = 'ERROR'
                    data = requests.get(self.user_urls['AddUserProperty1']).json()
                    message = data['msg']
                    property_id = data['data']['property_id']

            except ValueError:
                status_code = 'ERROR'
                property_id = None

            print(property_id, "user id, not property id")
            # self.property_ids.append(property_id)

            account_db_length_new = len(MetaClass)
            if account_db_length_new == account_db_length + 1:
                user_add = 'USER ADDED'
            else:
                user_add = 'USER NOT ADDED'

            new_api_DBreplacement_status = self.check_realtor_key_presence(time='a')

            if new_api_DBreplacement_status != api_DBreplacement_status:
                api_key = "API KEY REPLACED"
            else:
                api_key = "API KEY NOT REPLACED"

            unit_test_file.write(
                'Testing: {}; URL status(raw): {} URL status(interpretted): {} msg_line_endpoint: {} UserAdded?: {}, KeyChanged?: {}'.format(
                    'AddUserEndpoint',
                    status,
                    status_code,
                    message,
                    user_add,
                    api_key))
            unit_test_file.write('\n')

            # add a different property to see if the favorites list will change length by one
            favorite_length = MetaClass('uqowasfhsdfh', 'asdfjajfhksjhfkajs')
            try:
                status = requests.head(self.user_urls['AddUserProperty2']).status_code
                if 200 <= status < 300:
                    status_code = 'GOOD'
                    data = requests.get(self.user_urls['AddUserProperty2']).json()
                    message = data['msg']
                    property_id = data['data']['property_id']
                else:
                    status_code = 'ERROR'
                    data = requests.get(self.user_urls['AddUserProperty2']).json()
                    message = data['msg']
                    property_id = data['data']['property_id']
            except ValueError:
                property_id = None
                status_code = 'ERROR'

            # add the property id to the property_id list in the init method to use later for deletion
            # self.property_ids.append(property_id)

            # check if the property has been updated
            new_favorite_length = MetaClass('uqowasfhsdfh', 'asdfjajfhksjhfkajs')
            if new_favorite_length == favorite_length + 1:
                prop_add = 'PROP ADDED'
            else:
                prop_add = 'PROP NOT ADDED'

            unit_test_file.write(
                'Testing: {}; URL status(raw): {} URL status(interpretted): {} msg_line_endpoint: {} PropAdded?: {}'.format(
                    'AddUserDifferentProp',
                    status,
                    status_code,
                    message,
                    prop_add))
            unit_test_file.write('\n')

            # add the extra add user method tests here, these two fail, do not add anything
            response = self.AddUser_sameEmail(self.user_urls['AddUserSameEmailError'])
            unit_test_file.write(response)
            unit_test_file.write('\n')

            response = self.AddProp_dup(self.user_urls['AddUserProperty1'])
            unit_test_file.write(response)
            unit_test_file.write('\n')

            # this should create another user in the account db
            response = self.AddUser_SameNameDifEmail(self.user_urls['AddUserDifEmailSameName'])
            unit_test_file.write(response)
            unit_test_file.write('\n')

            # test the get_favorite_list here
            try:
                status = requests.head(self.user_urls['GetFavoriteList']).status_code
                if 200 <= status < 300:
                    status_code = 'GOOD'
                    data = requests.get(self.user_urls['GetFavoriteList']).json()
                    # print(data)
                    message = data['msg']
                    property_favs = data['properties']
                else:
                    status_code = 'ERROR'
                    data = requests.get(self.user_urls['GetFavoriteList']).json()
                    # print(data)
                    message = data['msg']
                    property_favs = data['properties']

            except ValueError:
                property_favs = []
                status_code = 'ERROR'

            # make sure the property favorites are of length 2
            favs_length = len(property_favs)
            if favs_length == 2:
                correct_length = "CORRECT LEN"
            else:
                correct_length = "INCORRECT LEN"

            unit_test_file.write(
                'Testing: {}; URL status(raw): {} URL status(interpretted): {} msg_line_endpoint: {} FavoriteLengthCorrect?: {}'.format(
                    'FavoriteListEndpoint',
                    status,
                    status_code,
                    message,
                    correct_length))

            unit_test_file.write('\n')

            # test the deletion urls here for each property_id in the arr in the init method
            self.property_ids.clear()
            self.get_property_ids([], user_id='primary')
            print(self.property_ids, type(self.property_ids))

            for prop_id in self.property_ids:
                print(prop_id)
                try:
                    status = requests.head(self.user_urls['DelProperty']).status_code
                    if 200 <= status < 300:
                        status_code = 'GOOD'
                        data = requests.get(self.user_urls['DelProperty'] + prop_id).json()
                        print(data)
                        # print(data)
                        message = data['msg']
                    else:
                        status_code = 'ERROR'
                        # print(self.user_urls['DelProperty'] + prop_id)
                        data = requests.get(self.user_urls['DelProperty'] + prop_id).json()
                        # print(data)
                        message = data['msg']

                except ValueError:
                    status_code = "ERROR"

            favorite_length = MetaClass('uqowasfhsdfh', 'asdfjajfhksjhfkajs')
            if favorite_length == 0:
                deleted_favorites = "SUCCESS"
            else:
                deleted_favorites = "FAILED"

            unit_test_file.write(
                'Testing: {}; URL status(raw): {} URL status(interpreted): {} msg_line_endpoint: {} PropAdded?: {}'.format(
                    'DelPropertyEndpoint',
                    status,
                    status_code,
                    message,
                    deleted_favorites))

            self.remove_added_test_users(user_opt=1)
            self.remove_added_test_users(user_opt=2)

            unit_test_file.write('\n')
            unit_test_file.close()

    def test_property_endpoints(self):
        with open('UnitTestFile.txt', 'a+') as unit_test_file:
            try:
                status = requests.head(self.property_urls['LoadCoordinates']).status_code
                if 200 <= status < 300:
                    status_code = 'GOOD'
                    data = requests.get(self.property_urls['LoadCoordinates']).json()
                    if isinstance(data['address_points'], list):
                        data_type = "LIST/PROPER"
                else:
                    status_code = 'ERROR'
                    data_type = 'NONLIST/IMPROPER'

            except ValueError:
                status_code = "ERROR"
                data_type = 'NONLIST/IMPROPER'

            unit_test_file.write(
                'Testing: {}; URL status(raw): {} URL status(interpretted): {} RadiusOutputType?: {}'.format(
                    'LoadRadiusCoordinates',
                    status,
                    status_code,
                    data_type))
            unit_test_file.write('\n')

            try:
                data = requests.get(self.property_urls['GetPropertyDetails']).json()
                if data['code'] == 200:
                    status_code = 'GOOD'
                    message = data['msg']
                    if len(data.keys()) > 0:
                        property_returned = "DETAILS PRESENT"
                    else:
                        property_returned = "DETAILS NULL TYPE"
                else:
                    status_code = 'ERROR'
                    message = 'no details for this property'
                    property_returned = None

            except ValueError:
                status_code = 'ERROR'
                message = 'no details for this property'
                property_returned = None

            unit_test_file.write(
                'Testing: {}; URL status(raw): {} URL status(interpretted): {} message_returned: {} PropertyReturned?: {}'.format(
                    'PropertyDetailsEndpoint',
                    status_code,
                    status_code,
                    message,
                    property_returned))

            unit_test_file.write('\n')
            unit_test_file.close()

    @time_function
    def main(self, *args, **kwargs):
        if len(args) == 0 or len(kwargs.keys()) == 0:
            self.test_user_endpoints()
            self.test_property_endpoints()
            print("done here")
        else:
            raise Exception('Uncalled for Arguments')


class PeekAPIEndpointTest(SetUpDBS):

    def __init__(self, dataset, batch_size):
        self.dataset = dataset
        self.batch_size = batch_size
        super(PeekAPIEndpointTest, self).__init__(self.dataset, 50)

        self.additional_urls = {
            'load_radius': 'http://127.0.0.1:5000/load_properties_api/?latitude=43.092541&longitude=-77.640418&radius=100',
            'polygon_search': 'http://127.0.0.1:5000/polygon_search/?edges=[[43.0863315, -77.6400726], [43.0863315, -77.6400726], [43.0824877, -77.6338292], [43.0841447, -77.639539]]',
        }

        self.reverse_geocoding_urls = {
            'reverse_single': 'http://127.0.0.1:5000/reverse_geocode/?coordinate=[43.1017166, -77.6342237]&radius=100',
            'batch_reverse': 'http://127.0.0.1:5000/batch_reverse/?coordinates=[[43.1018722, -77.6334654], [43.09245, -77.6353749], [43.0875773, -77.6383543], [43.0899153, -77.6424486], [43.0876896, -77.6440264], [43.0913689, -77.6467333], [43.0910046, -77.6455201]]&radius=10'
        }

        self.geocoding_urls = {
            'geocode_single': 'http://127.0.0.1:5000/geocoding/?address=20 Manor House Lane, Dobbs Ferry, NY, 10522&radius=10',
            'batch_geocode': 'http://127.0.0.1:5000/batch_geocoding/?addresses=["20 Manor House Lane, Dobbs Ferry, NY, 10522", "18 Manor House Lane, Dobbs Ferry, NY, 10522"]&radius=100'
        }

    def test_api(self):
        with open('UnitTestFile.txt', 'a+') as unit_test_file:
            unit_test_file.write('--------- Peek API test --------- :{}'.format(datetime.datetime.now()))
            unit_test_file.write('\n')

            # load_radius search here
            try:
                status = requests.head(self.additional_urls['load_radius']).status_code
                if 200 <= status < 300:
                    data = requests.get(self.additional_urls['load_radius']).json()

                    # there should be 8 properties returned
                    build = data['meta']['build']
                    radius_response_length = len(data['meta']['geocoded_data']['radius_array'])
                    if radius_response_length == 9:
                        message = 'RESPONSE SUCCESS'
                    else:
                        message = 'RESPONSE SUCCESS/SHORT!'

                else:
                    build = 0
                    radius_response_length = 0
                    message = 'RESPONSE FAILURE'

            except ValueError:
                build = 0
                radius_response_length = 0
                message = 'RESPONSE FAILURE'

            unit_test_file.write(
                'Testing: {}; URL status(raw): {} URL status(interpretted): {} build_id: {} RadiusResponseLength?: {}'.format(
                    'LoadRadiusEndpoint',
                    status,
                    message,
                    build,
                    radius_response_length))
            unit_test_file.write('\n')

            # polygon search test here
            try:
                status = requests.head(self.additional_urls['polygon_search']).status_code
                if 200 <= status < 300:
                    data = requests.get(self.additional_urls['polygon_search']).json()

                    # there should be 8 properties returned
                    radius_response_length = len(data['data'])
                    if radius_response_length == 8:
                        message = 'RESPONSE SUCCESS'
                    else:
                        message = 'RESPONSE SUCCESS/SHORT!'

                else:
                    radius_response_length = 0
                    message = 'RESPONSE FAILURE'

            except ValueError:
                radius_response_length = 0
                message = 'RESPONSE FAILURE'

            unit_test_file.write(
                'Testing: {}; URL status(raw): {} URL status(interpretted): {} build_id: {} ConvexResponseLength?: {}'.format(
                    'ConvexLocationEndpoints',
                    status,
                    message,
                    build,
                    radius_response_length))
            unit_test_file.write('\n')

            # reverse geocode (singular)
            try:
                status = requests.head(self.reverse_geocoding_urls['reverse_single']).status_code
                if 200 <= status < 300:
                    data = requests.get(self.reverse_geocoding_urls['reverse_single']).json()

                    # there should be 8 properties returned
                    build = data['meta']['build']
                    # print(data)
                    address = data['meta']['geocoded_data']['address_point']
                    if address == '88 Crittenden Way, Unit 2, Brighton, 14623, NY':
                        message = 'RESPONSE SUCCESS'
                    else:
                        address = None
                        build = 0
                        message = 'INCORRECT ADDRESS CONVERSION ERROR'

                else:
                    build = 0
                    address = None
                    message = 'RESPONSE FAILURE'

            except ValueError:
                build = 0
                radius_response_length = 0
                message = 'RESPONSE FAILURE'

            unit_test_file.write(
                'Testing: {}; URL status(raw): {} URL status(interpretted): {} build_id: {} SingularReverseGeocoder?: {}'.format(
                    'Reverse Geocoding Singular',
                    status,
                    message,
                    build,
                    address))
            unit_test_file.write('\n')

            # batch reverse geocoding test here ------------------------
            try:
                status = requests.head(self.reverse_geocoding_urls['batch_reverse']).status_code
                if 200 <= status < 300:
                    data = requests.get(self.reverse_geocoding_urls['batch_reverse']).json()

                    # there should be 8 properties returned
                    print(data)
                    build = data['meta']['build']
                    address_length = len(data['meta']['reverse_geocoded_data']['address_point'])
                    print(address_length)
                    if address_length == 5:
                        message = 'RESPONSE SUCCESS'
                    else:
                        address_length = 0
                        build = 0
                        message = 'NOT ENOUGH ADDRESSES'

                else:
                    build = 0
                    address_length = 0
                    message = 'RESPONSE FAILURE'

            except ValueError:
                build = 0
                radius_response_length = 0
                message = 'RESPONSE FAILURE'

            unit_test_file.write(
                'Testing: {}; URL status(raw): {} URL status(interpretted): {} build_id: {} BatchReverseGeocoder?: {}'.format(
                    'Batch Reverse Geocoding',
                    status,
                    message,
                    build,
                    address_length))
            unit_test_file.write('\n')

            # geocoding endpoint
            try:
                status = requests.head(self.geocoding_urls['geocode_single']).status_code
                if 200 <= status < 300:
                    data = requests.get(self.geocoding_urls['geocode_single']).json()

                    print(data)
                    coordinates = data['Valid Response'][0]['Coordinates']
                    if coordinates == [41.0106378, -73.8619343]:
                        message = 'RESPONSE SUCCESS'
                    else:
                        build = 0
                        message = 'INCORRECT COORDINATES'

                else:
                    build = 0
                    coordinates = [None, None]
                    message = 'RESPONSE FAILURE'

            except ValueError:
                build = 0
                coordinates = [None, None]
                message = 'RESPONSE FAILURE'

            unit_test_file.write(
                'Testing: {}; URL status(raw): {} URL status(interpretted): {} build_id: {} CoordinatesReturned?: {}'.format(
                    'geocoding singular',
                    status,
                    message,
                    build,
                    coordinates))
            unit_test_file.write('\n')

            # batch geocoder endpoint
            try:
                status = requests.head(self.geocoding_urls['batch_geocode']).status_code
                if 200 <= status < 300:
                    data = requests.get(self.geocoding_urls['batch_geocode']).json()

                    print(data)
                    coordinates = len(data['Coordinates'])
                    if coordinates == 7:
                        message = 'RESPONSE SUCCESS'
                    else:
                        build = 0
                        message = 'INCORRECT COORDINATES'

                else:
                    build = 0
                    coordinates = [None, None]
                    message = 'RESPONSE FAILURE'

            except ValueError:
                build = 0
                coordinates = [None, None]
                message = 'RESPONSE FAILURE'

            unit_test_file.write(
                'Testing: {}; URL status(raw): {} URL status(interpretted): {} build_id: {} CoordinatesReturned?: {}'.format(
                    'batch Geocoding Singular',
                    status,
                    message,
                    build,
                    coordinates))

            unit_test_file.write('\n')
            unit_test_file.write('\n')
            unit_test_file.write('\n')
            unit_test_file.close()

    # @time_function
    def main(self):
        self.test_api()


def main(dataset, batch_size):
    Peek_mobile_test = PeekAppEndpointTest(dataset)
    Peek_api_test = PeekAPIEndpointTest(dataset, batch_size)

    Peek_mobile_test.main()
    Peek_api_test.main()


#main('PeekDB', 50)

def test_endpoint_runtime():
    runtimes = []
    url = 'http://127.0.0.1:8000/load_properties_api/parcel?latitude=43.092541&longitude=-77.640418&radius=200&reverse_param=false'
    for _ in range(10):
        start = time.perf_counter()
        data = requests.get(url)
        print(data)
        runtime = time.perf_counter() - start
        runtimes.append(runtime)
    print(sum(runtimes)/len(runtimes))

#test_endpoint_runtime()



from pydantic import BaseModel, validator
from typing import List, Optional, Union, Any, Tuple, Callable
import sys
from abc import ABC, abstractmethod
import places_geocode.app.generator_BSON as generator
from places_geocode.app.models import common_models



# noinspection PyCompatibility
class DensitySchema(BaseModel):
    """base model for density endpoint data"""
    units: Optional[List[str]] = ['km', 'mi', 'm', 'ft', 'yd']
    radius: Optional[Union[int, float]]
    region_area: Optional[Union[int, float]] = 0
    iterative_counter: Optional[int] = 0
    query: Optional[List[Any]] = []

    @validator("radius")
    @classmethod
    def validate_radius(cls, radius):
        """validate the radius inputted"""
        if not (isinstance(radius, int) or isinstance(radius, float)):
            raise Exception("Radius Datatype Incorrect")


def convert_generator_exp(mongo_obj: Any, generator_strategy: Callable, parser: Callable) -> List[Any]:
    """convert the generator"""
    query_arr = generator.handler_func(mongo_obj, generator_strategy(), parser, option=0)
    return query_arr


# Density (Custom) strategies below
class DensityStrategy(ABC):
    """abstract base class custom density strategies"""

    @abstractmethod
    def create_dense_strategy(self, collection, level: Union[str, None], generator_strategy: Callable, parser: Callable, *es) -> Tuple[List, int]:
        """create the density strategy"""
        pass


class StateDensityStrategy(DensityStrategy):
    """State strategy for custom density endpoints"""

    def create_dense_strategy(self, collection, level: Union[int, str, None], generator_strategy: Callable, parser: Callable, *es) -> Tuple[
        List, int]:
        """create the density strategy"""

        query = collection.find({
            'properties.region': level
        })
        query_arr = convert_generator_exp(query, generator_strategy, parser)
        try:
            length = len(query_arr)
        except TypeError:
            length = 0
        return query_arr, length

    @validator("level")
    @classmethod
    def _check_level(cls, level):
        if level not in CommonUtilityModel.states:
            raise Exception("State not Found")


class PostcodeDensityStrategy(DensityStrategy):
    """postcode strategy for density computation"""

    def create_dense_strategy(self, collection, level: Union[int, str, None], generator_strategy: Callable, parser: Callable, *es) -> Tuple[
        List, int]:
        """create the density strategy"""
        query = collection.find({
            'properties.postcode': str(level)
        })
        query_arr = convert_generator_exp(query, generator_strategy, parser)
        try:
            length = len(query_arr)
        except TypeError:
            length = 0
        return query_arr, length

    @validator("level")
    @classmethod
    def _check_level(cls, level):
        if len(level) != 5:
            raise Exception("Invalid Postcode Entered")


class CityDensityStrategy(DensityStrategy):
    """city density strategy for city approximation of parcel density"""

    def create_dense_strategy(self, collection, level: Union[int, str, None], generator_strategy: Callable, parser, *es) -> Tuple[
        List, int]:
        """create the density strategy"""
        query = collection.find({
            'properties.city': level
        })
        query_arr = convert_generator_exp(query, generator_strategy, parser)
        try:
            length = len(query_arr)
        except TypeError:
            length = 0
        return query_arr, length


def conduct_elastic_query(connection: Any, index: str, body: dict) -> list:
    """run the elastic query"""

    result = connection.search(index=index, body=body)['hits']['hits']
    return result


class DensityStrategyElastic(ABC):
    """density model class for elasticsearch"""

    @abstractmethod
    def create_dense_strategy(self, index: str, level: Union[str, None], generator_strategy: Callable, parser: Callable, *es) -> Tuple[
        List, int]:
        """create the density strategy involving elasticsearch implementing encapsulation design pattern"""
        pass


class StateDensityElastic(DensityStrategyElastic):
    """find the density of the state"""

    def create_dense_strategy(self, index: str, level: Union[str, None], generator_strategy: Callable, parser: Callable, *es) -> Tuple[List, int]:
        """make elastic density strategy for state-level query"""

        # create the query body input
        body = {
            "query": {
                "bool": {
                    "should": [
                        {"term": {"properties.region": level}}
                    ]
                }
            }
        }

        results = conduct_elastic_query(es[0], index, body)

        length = len(results)
        query_arr = convert_generator_exp(results, generator_strategy, parser)

        return query_arr, length

    @validator("level")
    @classmethod
    def _check_level(cls, level):
        if level not in common_models.CommonUtilityModel.states:
            raise Exception("State not Found")


class PostcodeDensityElastic(DensityStrategyElastic):
    """create postcode strategy option for elasticsearch"""

    def create_dense_strategy(self, index: str, level: Union[str, None], generator_strategy: Callable, parser: Callable, *es) -> Tuple[
        List, int]:
        """make density elastic pattern for postcodes"""

        body = {
            "query": {
                "bool": {
                    "should": [
                        {"term": {"properties.postcode": level}}
                    ]
                }
            }
        }

        # compute elasticsearch query
        results = conduct_elastic_query(es[0], index, body)

        try:
            length = len(results)
        except TypeError:
            length = 0

        # convert the array to coordinates
        query_arr = convert_generator_exp(results, generator_strategy, parser)

        return query_arr, length

    @validator("level")
    @classmethod
    def _check_level(cls, level):
        if len(level) != 5:
            raise Exception("Invalid Postcode Entered")


class CityDensityElastic(DensityStrategyElastic):
    """create elastic strategy for city density query"""

    def create_dense_strategy(self, index: str, level: Union[str, None], generator_strategy: Callable, parser: Callable, *es) -> Tuple[List, int]:
        """create the density strategy for general city queries"""

        # create the body of the query outgoing request
        body = {
            "query": {
                "fuzzy": {
                    "properties.city": {
                        "value": level,
                        "fuzziness": "AUTO",
                        "max_expansions": 50,
                        "prefix_length": 0,
                        "transpositions": True,
                        "rewrite": "constant_score"
                    }
                }
            }
        }

        results = conduct_elastic_query(es[0], index, body)

        try:
            length = len(results)
        except TypeError:
            length = 0

        query_arr = convert_generator_exp(results, generator_strategy, parser)

        return query_arr, length


def strategies_main(collection: Any, query_value: Union[int, str, None], strategy, generator_strategy: Callable, parser: Callable,
                    *es) -> Tuple:
    """return the correct strategy"""

    # test for presence of elastic request
    try:
        # elasticsearch option
        es = es[0]
        results = strategy.create_dense_strategy(collection, query_value, generator_strategy, parser, es)
    except IndexError:
        # pymongo option
        results = strategy.create_dense_strategy(collection, query_value, generator_strategy, parser)

    # send results to client code
    return results
